// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | info@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// QML component is used to specify single spectral line options to be used in calculations
import QtQuick 2.4
import QtQuick.Controls 1.3
import QtQuick.Window 2.2
import QtQuick.Dialogs 1.2
import QtQuick.Layouts 1.1

Item {

    id: lineOptionsItem
    anchors.fill: parent

    // Rectangle to draw the form border
    Rectangle {
        anchors.fill: parent
        border.width: 1
        border.color: "black"
        color: Qt.rgba(0, 0, 0, 0.6)

        // This is introduced to disable items located on the background
        MouseArea {
            anchors.fill: parent
            onClicked: { }
        }
    }

    // Rectangle to draw the form border
    Rectangle {
        anchors.right: mainLineOptionsColumnLayout.right
        anchors.top: mainLineOptionsColumnLayout.top
        anchors.left: mainLineOptionsColumnLayout.left
        anchors.bottom: mainLineOptionsColumnLayout.bottom
        anchors.margins: ((Qt.platform.os == 'ios') || (Qt.platform.os == 'android') || (Qt.platform.os == 'winrt') || (Qt.platform.os == 'winphone')) ? lineShapeLabel.height * (-0.85) : -20
        color: "white"
        border.width: 1
        border.color: "black"
        radius: height * 0.02
    }

    // Button to close the form
    BackRectangle {
        anchors.fill: exitButton
        anchors.margins: 8
    }
    ControlButton {
        anchors.right: lineOptionsItem.right
        anchors.top: lineOptionsItem.top
        id: exitButton
        platformText: qsTr("X")
        opacity: 1.0

        // Enabling the closing of the form when the exit button is pressed
        MouseArea {
            anchors.fill: parent
            onClicked: {
                lineOptionsItem.destroy();
            }
        }
    }

    // Column layout to hold all the interface items
    ColumnLayout {
        // At rotation the buttons are not positioned exactly the way they should be (fix this)
        id: mainLineOptionsColumnLayout
        anchors.centerIn: parent
        spacing: 5
        transformOrigin: Item.Center
        Layout.fillWidth: true
        Layout.fillHeight: true

        PlatformLabel {
            Layout.fillWidth: true
            id: commentLabel
            text: qsTr("Single spectral line calculation options.")
        }

        GridLayout {
            rows: 5
            columns: 2

            PlatformLabel {
                id: lineShapeLabel
                anchors.right: lineShapeComboBox.left
                Layout.column: 0
                Layout.row: 0
                text: qsTr("Lineshape profile:  ")
            }

            // ComboBox to select the lineshape profile
            PlatformComboBox {
                id: lineShapeComboBox
                Layout.fillWidth: true
                Layout.column: 1
                Layout.row: 0
                implicitWidth: lineShapeLabel.width
                model: [ "Doppler", "Lorentz", "Voigt / Standard precision", "Voigt / Higher precision"]
                currentIndex: lineOptDataObj.lineshapeProfile
                onCurrentIndexChanged: lineOptDataObj.lineshapeProfile = currentIndex
                style: ((Qt.platform.os == 'winrt') || (Qt.platform.os == 'winphone')) ? winReferenceComboBox.style : referenceComboBox.style
            }

            PlatformLabel {
                id: intensityThresholdLabel
                anchors.right: intensityThresholdTextField.left
                Layout.column: 0
                Layout.row: 1
                text: qsTr("   Intensity threshold:\n(cm * molec * 1E-25)  ")
            }

            // Changing the Intensity threshold
            PlatformTextField {
                id: intensityThresholdTextField
                Layout.fillWidth: true
                //anchors.top: intensityThresholdLabel.top
                Layout.column: 1
                Layout.row: 1
                implicitWidth: lineShapeLabel.width
                text: lineOptDataObj.intensThresh
                onTextChanged: lineOptDataObj.intensThresh = text
            }

            PlatformLabel {
                id: wingContribSatHwLabel
                anchors.right: wingContribSatHwTextField.left
                Layout.column: 0
                Layout.row: 2
                text: qsTr(" Line wing contribution:\n  (saturated halfwidths)  ")
            }

            // Changing the line wing contribution in halfwidths
            PlatformTextField {
                id: wingContribSatHwTextField
                //anchors.bottom: wingContribSatHwLabel.bottom
                Layout.fillWidth: true
                Layout.column: 1
                Layout.row: 2
                implicitWidth: lineShapeLabel.width
                text: lineOptDataObj.wingContribSatHw
                onTextChanged: lineOptDataObj.wingContribSatHw = text
            }

            PlatformLabel {
                id: wingContribAbsLabel
                anchors.right: wingContribAbsTextField.left
                Layout.column: 0
                Layout.row: 3
                text: qsTr("Line wing contribution:\n       (absolute, in cm-1)  ")
            }

            // Changing the line wing contribution absolute
            PlatformTextField {
                id: wingContribAbsTextField
                //anchors.bottom: wingContribAbsLabel.bottom
                Layout.fillWidth: true
                Layout.column: 1
                Layout.row: 3
                implicitWidth: lineShapeLabel.width
                text: lineOptDataObj.wingContribAbs
                onTextChanged: lineOptDataObj.wingContribAbs = text
            }

            // Label corresponding to the broadening type selection
            PlatformLabel {
                id: broadeningLabel
                anchors.right: broadeningComboBox.left
                Layout.column: 0
                Layout.row: 4
                text: qsTr("Broadening type:  ")
            }

            // ComboBox to select the Broadening type
            PlatformComboBox {
                id: broadeningComboBox
                Layout.fillWidth: true
                Layout.column: 1
                Layout.row: 4
                implicitWidth: lineShapeLabel.width
                model: [ "Air", "Self", "Air and Self"]
                currentIndex: lineOptDataObj.lineBroadening
                onCurrentIndexChanged: lineOptDataObj.lineBroadening = currentIndex
                style: ((Qt.platform.os == 'winrt') || (Qt.platform.os == 'winphone')) ? winReferenceComboBox.style : referenceComboBox.style
            }
            // This weird way was introduced to enable text black color under the windows phone platform while maintaining the style a default for other platforms
            PlatformComboBox {
                id: referenceComboBox
                visible: false
            }
            PlatformComboBox {
                id: winReferenceComboBox
                style: ComboBoxPlatformStyle {}
                visible: false
            }

            PlatformLabel {
                id: useLineShiftLabel
                anchors.right: useLineShiftSwitch.left
                Layout.column: 0
                Layout.row: 5
                text: qsTr("Use line shift:  ")
            }

            // ComboBox to select line shift
            Switch {
                id: useLineShiftSwitch
                Layout.column: 1
                Layout.row: 5
                implicitWidth: useLineShiftLabel.width * 0.85
                checked: lineOptDataObj.useLineShift
                onCheckedChanged: lineOptDataObj.useLineShift = checked
                height: useLineShiftLabel.height * 0.75
            }
        }
    }
}

