// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | info@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// Main window contains the plot and the main navigation items
import QtQuick 2.4
import QtQuick.Controls 1.3
import QtQuick.Layouts 1.1
import Plot 1.0
import QtQuick.Controls.Styles 1.4

Item {

    id: mainWindowItem

    property alias menuButton: menuButton
    property alias calcButton: calcButton
    property alias exitButton: exitButton
    property alias moleculesButton: moleculesButton
    property alias spectrumPlot: spectrumPlot
    property alias plotAndMolViewRect: plotAndMolViewRect
    property alias totalMolsButton: totalMolsButton
    property alias menuButtonText: menuButtonText
    property alias molFlickable: molFlickable  // This line should be commented if the scrollview is disabled (for Windows Phone application)
    property alias totalsButtonRect: totalsButtonRect
    property alias plotPinchArea: plotPinchArea
    property alias showTotalOptionsRectangle: showTotalOptionsRectangle
    property alias showTotalOrRadianceButton: showTotalOrRadianceButton

    GridLayout {
        id: mainGridLayout
        anchors.fill: parent
        rowSpacing: 5
        transformOrigin: Item.Center

        // Rectangle to contain the calculated plot
        Rectangle {
            id: plotAndMolViewRect
            anchors.right: mainGridLayout.right
            anchors.top: mainGridLayout.top
            anchors.left: mainGridLayout.left
            anchors.bottom: mainGridLayout.bottom
            color: "white"
            border.width: 1
            border.color: "white"
            radius: 3

            // Plot to contain the plotted spectra (will also be used to manipulate (scaling, resizing etc.) the spectra through touch etc.)
            Plot {
                id: spectrumPlot
                anchors.fill: parent
                xMinPadding: menuButton.width * 0.87
                xMaxPadding: menuButton.width * 1.02
                yMinPadding: (Qt.platform.os == 'ios') ? menuButton.height + 2 : menuButton.height
                yMaxPadding: menuButton.height * 1.23

                yAxisTitle: specResultsDataObj.yAxisTitle
                xAxisTitle: specResultsDataObj.xAxisTitle

                minX: specResultsDataObj.minX
                maxX: specResultsDataObj.maxX
                minY: specResultsDataObj.minY
                maxY: specResultsDataObj.maxY

                multiplier: specResultsDataObj.multiplier

                spectrum: specResultsDataObj.spectrum
                wnGrid: specResultsDataObj.wnGrid                
                lineColor: specResultsDataObj.lineColor
                plotInverted: false
                lineThickness: 1

                currSpecIndex: specResultsDataObj.currSpecIndex

                numMajorXTicks: 5
                numMinorXTicks: 5
                numMajorYTicks: 5
                numMinorYTicks: 5

                saveImage: shareResultsDataObj.saveImage;                
                topString: shareResultsDataObj.topString;
                fileWritingComplete: shareResultsDataObj.fileWritingComplete;

                // Used to enable zoom of the plot and other functions
                PinchArea {
                    id: plotPinchArea
                    anchors.fill: parent
                }
            }

            Rectangle {
                id: totalsButtonRect
                height: totalMolsButton.height * 1.15
                anchors.bottom: plotAndMolViewRect.bottom
                anchors.bottomMargin: menuButton.height * 1.22
                anchors.right: plotAndMolViewRect.right
                color: 'white'
                width: molFlickable.width
            }

            // Flickable is used to add a scroll bar when the number of items is larger than the screen can fit
            // This scroll view needs to be removed when compiling under Windows Phone and Windows RT, otherwise the scroll does not work
            // Under other platforms the scroll keeps the items in the list within the desired boundaries so should be kept
            ScrollView {
                id: molFlickable
                anchors.top: plotAndMolViewRect.top
                anchors.topMargin: (Qt.platform.os == 'ios') ? menuButton.height * 1.02 + 2 : menuButton.height * 1.02
                anchors.right: plotAndMolViewRect.right
                anchors.bottom: totalsButtonRect.top
                verticalScrollBarPolicy: Qt.ScrollBarAlwaysOff
                implicitWidth: menuButton.width
                enabled: true

                ListView {

                    id: molListView
                    spacing: 0
                    model: molListViewReducedModel
                    visible: !specResultsDataObj.showTotal

                    // This should be enabled if not under Windows Phone (couldn't find a better solution for now)
                    anchors.fill: parent

//                    // This should be uncommented under Windows Phone (couldn't find a better solution for now), otherwise the scroll will not work under Windows Phone
//                    anchors.top: plotAndMolViewRect.top
//                    anchors.topMargin: (Qt.platform.os == 'ios') ? menuButton.height * 1.02 + 2 : menuButton.height * 1.02
//                    anchors.right: plotAndMolViewRect.right
//                    anchors.bottom: totalsButtonRect.top
//                    implicitWidth: menuButton.width
//                    enabled: true

                    delegate: RowLayout {

                        id: moleculeDelegate
                        implicitWidth: menuButton.width

                        RowLayout {
                            id: molColorRowLayout

                            // Rectangle to draw the color to be plotted
                            Rectangle {
                                width: menuButton.width
                                height: indexText.height
                                border.width: 1
                                border.color: model.color
                                color: model.color
                                radius: 1
                                Text {
                                    id: indexText
                                    text: " " + model.molName
                                    font.bold: true
                                    horizontalAlignment: Text.AlignRight
                                    color: 'white'
                                }
                            }
                        }
                    }
                }
            }
        }

        // Menu buttons size has been minimized to enlarge the area available for the plot
        Button {
            anchors.left: mainGridLayout.left
            anchors.top: mainGridLayout.top
            anchors.leftMargin: menuButton.width * 0.87
            anchors.margins: (Qt.platform.os == 'ios') ? 1 : 0
            id: menuButton
            Text {
                anchors.centerIn: parent
                id: menuButtonText
                text: qsTr("Menu")
            }
            implicitHeight: menuButtonText.height * 2
            implicitWidth: menuButtonText.width * 1.5
        }
        Button {
            anchors.left: menuButton.right
            anchors.top: mainGridLayout.top
            anchors.leftMargin: menuButtonText.width * 0.5
            anchors.margins: (Qt.platform.os == 'ios') ? 1 : 0
            id: calcButton
            Text {
                anchors.centerIn: parent
                id: calcButtonText
                text: qsTr("Calc")
            }
            implicitHeight: menuButton.height
            implicitWidth: menuButton.width
        }
        Button {
            anchors.right: mainGridLayout.right
            anchors.top: mainGridLayout.top
            anchors.margins: (Qt.platform.os == 'ios') ? 1 : 0
            id: exitButton
            Text {
                anchors.centerIn: parent
                id: exitButtonText
                text: qsTr("Exit")
            }
            implicitHeight: menuButton.height
            implicitWidth: menuButton.width
        }
        Button {
            anchors.left: calcButton.right
            anchors.top: mainGridLayout.top
            anchors.leftMargin: menuButtonText.width * 0.5
            anchors.margins: (Qt.platform.os == 'ios') ? 1 : 0
            id: moleculesButton
            Text {
                anchors.centerIn: parent
                id: molsButtonText
                text: qsTr("Mols")
            }
            implicitHeight: menuButton.height
            implicitWidth: menuButton.width
        }

        Rectangle {
            id: showTotalOptionsRectangle
            anchors.top: exitButton.bottom
            anchors.bottom: totalMolsButton.top
            anchors.right: mainGridLayout.right
            anchors.left: exitButton.left
            anchors.margins: exitButton.width * 0.05
            visible: false

            ColumnLayout {
                id: showTotalOptionsColumnLayout
                anchors.centerIn: parent

                Button {
                    id: showPlainTotalButton
                    anchors.right: showTotalOptionsColumnLayout.right
                    anchors.left: showTotalOptionsColumnLayout.left
                    anchors.margins: 1
                    checkable: true
                    onCheckedChanged: specResultsDataObj.showPlainTotal = checked;
                    checked: specResultsDataObj.showPlainTotal
                    implicitHeight: menuButton.width * 0.95 - 3
                    implicitWidth: menuButton.width * 0.95
                    style: ButtonStyle
                    {
                        background: Rectangle {
                            border.color: 'black'
                            border.width: 2
                            color: control.checked ? 'black' : 'white'
                            radius: height * 0.5
                        }
                    }
                }

                Button {
                    id: showInstrTotalButton
                    anchors.right: showTotalOptionsColumnLayout.right
                    anchors.left: showTotalOptionsColumnLayout.left
                    anchors.margins: 2
                    checkable: true
                    onCheckedChanged: specResultsDataObj.showInstrTotal = checked;
                    checked: specResultsDataObj.showInstrTotal
                    implicitHeight: menuButton.width * 0.95 - 3
                    implicitWidth: menuButton.width * 0.95
                    style: ButtonStyle
                    {
                        background: Rectangle {
                            border.color: 'red'
                            border.width: 2
                            color: control.checked ? 'red' : 'white'
                            radius: height * 0.5
                        }
                    }
                }

                Button {
                    id: showTotalOrRadianceButton
                    anchors.right: showTotalOptionsColumnLayout.right
                    anchors.left: showTotalOptionsColumnLayout.left
                    anchors.margins: 2
                    checkable: true
                    onCheckedChanged: specResultsDataObj.showRadiance = checked
                    checked: specResultsDataObj.showRadiance
                    implicitHeight: menuButton.width * 0.95 - 3
                    implicitWidth: menuButton.width * 0.95

                    Text {
                        text: "R"
                        anchors.centerIn: parent
                        color: showTotalOrRadianceButton.checked ? 'white' : 'grey'
                    }

                    style: ButtonStyle
                    {
                        background: Rectangle {
                            id: auxRect
                            border.color: 'grey'
                            border.width: 2
                            color: control.checked ? 'grey' : 'white'
                            radius: height * 0.5
                        }                        
                    }
                }
            }
        }

        Button {
            id: totalMolsButton
            anchors.bottom: mainGridLayout.bottom
            anchors.right: mainGridLayout.right
            anchors.rightMargin: 1
            anchors.bottomMargin: menuButton.height * 1.15
            checkable: true
            onCheckedChanged: specResultsDataObj.showTotal = checked;
            checked: specResultsDataObj.showTotal
            clip: true

            Text {
                anchors.centerIn: parent
                id: totalMolsButtonText
                text: qsTr("Total")
            }
            implicitHeight: menuButton.height
            implicitWidth: menuButton.width - 1
        }
    }
}
