// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | info@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// Parts of the line-by-line calculation algorithm are based on the Python code from the HITRAN Application Programming Interface (HAPI)
// R.V. Kochanov, I.E. Gordon, L.S. Rothman, P. Wcisło, C. Hill, J.S. Wilzewski, "HITRAN Application Programming Interface (HAPI):
// A comprehensive approach to working with spectroscopic data", JQSRT, Volume 177, July 2016, Pages 15–30; (hitran.org/hapi/)
//
// The Voigt calculation routine uses the algorithm described in: Abrarov, S., and Quine, B. (2015).
// "A rational approximation for efficient computation of the Voigt function in quantitative spectroscopy".
// Journal of Mathematics Research, 7(2), p 163. doi:http://dx.doi.org/10.5539/jmr.v7n2.
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// Object with all line-by-line calculation elements
#include <math.h>
#include <complex>
#include <QVector>
#include <QMessageBox>
#include <QApplication>
#include <QFile>
#include <QTextStream>
#include <QApplication>
#include "lbl.h"
#include "auxiliary.h"
#include "hitranonerec.h"
#include "const.h"


Lbl::Lbl(QObject *parent, HitranOnlineDownloader *newHitran) : QObject(parent)
{
    // Passing in a reference to the HITRAN download object
    this->hitran = newHitran;

    // The following coefficents used in the the Voigt profile implementation are taken from the publication reported in:
    // Abrarov, S., & Quine, B. (2015). A Rational Approximation for Efficient Computation of the Voigt Function
    // in Quantitative Spectroscopy. Journal Of Mathematics Research, 7(2), p163. doi:http://dx.doi.org/10.5539/jmr.v7n2
    coeff1_12.resize(12);
    coeff2_12.resize(12);
    coeff3_12.resize(12);
    coeff1_12[0] = 2.307372754308023e-001; coeff2_12[0] = 4.989787261063716e-002; coeff3_12[0] = 1.464495070025765e+000;
    coeff1_12[1] = 7.760531995854886e-001; coeff2_12[1] = 4.490808534957343e-001; coeff3_12[1] = -3.230894193031240e-001;
    coeff1_12[2] = 4.235506885098250e-002; coeff2_12[2] = 1.247446815265929e+000; coeff3_12[2] = -5.397724160374686e-001;
    coeff1_12[3] = -2.340509255269456e-001; coeff2_12[3] = 2.444995757921221e+000; coeff3_12[3] = -6.547649406082363e-002;
    coeff1_12[4] = -4.557204758971222e-002; coeff2_12[4] = 4.041727681461610e+000; coeff3_12[4] = 2.411056013969393e-002;
    coeff1_12[5] = 5.043797125559205e-003; coeff2_12[5] = 6.037642585887094e+000; coeff3_12[5] = 4.001198804719684e-003;
    coeff1_12[6] = 1.180179737805654e-003; coeff2_12[6] = 8.432740471197681e+000; coeff3_12[6] = -5.387428751666454e-005;
    coeff1_12[7] = 1.754770213650354e-005; coeff2_12[7] = 1.122702133739336e+001; coeff3_12[7] = -2.451992671326258e-005;
    coeff1_12[8] = -3.325020499631893e-006; coeff2_12[8] = 1.442048518447414e+001; coeff3_12[8] = -5.400164289522879e-007;
    coeff1_12[9] = -9.375402319079375e-008; coeff2_12[9] = 1.801313201244001e+001; coeff3_12[9] = 1.771556420016014e-008;
    coeff1_12[10] = 8.034651067438904e-010; coeff2_12[10] = 2.200496182129099e+001; coeff3_12[10] = 4.940360170163906e-010;
    coeff1_12[11] = 3.355455275373310e-011; coeff2_12[11] = 2.639597461102705e+001; coeff3_12[11] = 5.674096644030151e-014;

    coeff1_16.resize(16);
    coeff2_16.resize(16);
    coeff3_16.resize(16);
    coeff1_16[0] = 1.608290174437121e-001; coeff2_16[0] = 3.855314219175531e-002; coeff3_16[0] = 1.366578214428949e+000;
    coeff1_16[1] = 6.885967427017463e-001; coeff2_16[1] = 3.469782797257978e-001; coeff3_16[1] = -5.742919588559361e-002;
    coeff1_16[2] = 2.651151642675390e-001; coeff2_16[2] = 9.638285547938826e-001; coeff3_16[2] = -5.709602545656873e-001;
    coeff1_16[3] = -2.050008245317253e-001; coeff2_16[3] = 1.889103967396010e+000; coeff3_16[3] = -2.011075414803758e-001;
    coeff1_16[4] = -1.274551644219086e-001; coeff2_16[4] = 3.122804517532180e+000; coeff3_16[4] = 1.069871368716704e-002;
    coeff1_16[5] = -1.134971805306579e-002; coeff2_16[5] = 4.664930205202391e+000; coeff3_16[5] = 1.468639542320982e-002;
    coeff1_16[6] = 4.201921570328543e-003; coeff2_16[6] = 6.515481030406647e+000; coeff3_16[6] = 1.816268776500938e-003;
    coeff1_16[7] = 8.084740485193432e-004; coeff2_16[7] = 8.674456993144942e+000; coeff3_16[7] = -6.875907999947567e-005;
    coeff1_16[8] = 1.946391440605860e-005; coeff2_16[8] = 1.114185809341728e+001; coeff3_16[8] = -2.327910355924500e-005;
    coeff1_16[9] = -4.132639863292073e-006; coeff2_16[9] = 1.391768433122366e+001; coeff3_16[9] = -1.004011418729134e-006;
    coeff1_16[10] = -2.656262492217795e-007; coeff2_16[10] = 1.700193570656409e+001; coeff3_16[10] = 2.304990232059197e-008;
    coeff1_16[11] = -1.524188131553777e-009; coeff2_16[11] = 2.039461221943855e+001; coeff3_16[11] = 2.275276345355270e-009;
    coeff1_16[12] = 2.239681784892829e-010; coeff2_16[12] = 2.409571386984707e+001; coeff3_16[12] = 3.383885053101652e-011;
    coeff1_16[13] = 4.939143128687883e-012; coeff2_16[13] = 2.810524065778962e+001; coeff3_16[13] = -4.398940326332977e-013;
    coeff1_16[14] = 4.692078138494072e-015; coeff2_16[14] = 3.242319258326621e+001; coeff3_16[14] = -1.405511706545786e-014;
    coeff1_16[15] = -2.512454984032184e-016; coeff2_16[15] = 3.704956964627684e+001; coeff3_16[15] = -3.954682293307548e-016;

    // Method to create actions
    createActions();
}

// Method to create connections
void Lbl::createActions()
{
    // Need to expand this for other profiles
    connect(this->hitran, SIGNAL(dataAvailable()), this, SLOT(calcWnSegment()));
    connect(this->hitran, SIGNAL(noMoreDataAvailable()), this, SLOT(postCalculation()));
}

// Method to get and set a pointer to the object with the calculation parameters
Params *Lbl::getParams() { return this->par; }
void Lbl::setParams(Params *newParams) { this->par = newParams; }

// Method to get and set an object with wavelengths
Wavelength *Lbl::getWavelength() { return this->wavelength; }
void Lbl::setWavelength(Wavelength *newWavelength) { this->wavelength = newWavelength; }

// Methods to get and set the HITRAN records to run calculations for
QVector<HitranOneRec> *Lbl::getHitranRecs() { return this->hitranRecs; }
void Lbl::setHitranRecs(QVector<HitranOneRec> *newHitranRecs) { this->hitranRecs = newHitranRecs; }

// Methods to get and set a pointer to the spectra array (this object is used to save the calculated spectra to)
OneLayerSpectra *Lbl::getHorizSpec() { return this->horizSpectra; }
void Lbl::setHorizSpec(OneLayerSpectra *newHorizSpectra) { this->horizSpectra = newHorizSpectra; }

// Method to pass the HITRAN downloader class reference to the object (used for segment reading of the data)
void Lbl::setHitranDownloader(HitranOnlineDownloader *newHitran) { this->hitran = newHitran; }

// Method is used to set a variable used to cancel the calculation
void Lbl::setCancelCalculation(bool newCancelCalculation)
{
    this->cancelCalculation = newCancelCalculation;
    QApplication::processEvents();
}

// Method to initialize the spectra (resize them according to the molparam.txt file information)
void Lbl::initHorizSpectra()
{
    int i;  // Loop variable
    int j;  // Loop variable
    int k;  // Loop variable
    QMessageBox msgBox;


    // Initializing the horizontal path spectra
    horizSpectra->getMol()->resize(par->getMol()->count());
    for (i = 0; i < par->getMol()->count(); i++)
    {
        // Resizing the spectra according to the number of isotopes in the set of spectra
        horizSpectra->getMolAt(i)->getSpec()->getY()->resize(wavelength->getWns()->count());
        horizSpectra->getMolAt(i)->getSpec()->getX()->resize(0);
        horizSpectra->getMolAt(i)->getIso()->resize(par->getMolAt(i)->getIso()->count());
        for (j = 0; j < horizSpectra->getMolAt(i)->getIso()->count(); j++)
        {
            horizSpectra->getMolAt(i)->getIsoAt(j)->getSpec()->getY()->resize(wavelength->getWns()->count());
            horizSpectra->getMolAt(i)->getIsoAt(j)->getSpec()->getX()->resize(0);
        }
    }

    // Filling the arrays with 0 for the composite spectrum
    horizSpectra->getComp()->getY()->resize(wavelength->getWns()->count());
    horizSpectra->getComp()->getX()->resize(0);
    for (i = 0; i < wavelength->getWns()->count(); i++)
    {
        horizSpectra->getComp()->setYAt(i, 0);
    }

    // Filling arrays with 0 for the molecular and isotope spectra
    for (i = 0; i < par->getMol()->count(); i++)
    {
        // Filling the arrays with 0 for the molecular spectra
        for (k = 0; k < wavelength->getWns()->count(); k++)
        {
            horizSpectra->getMolAt(i)->getSpec()->setYAt(k, 0);
        }

        // Here the same should be done for the isotope spectra later on (if used)
    }
}

// Method to find maximum values in the spectra
void Lbl::findMinMaxValuesInSpectra()
{
    int i;  // Loop variable


    // Filling the arrays with 0 for the composite spectrum
    horizSpectra->getComp()->setMaxValue(Auxiliary::maximum(horizSpectra->getComp()->getY()));
    // horizSpectra->getComp()->setMinValue(Auxiliary::minimum(horizSpectra->getComp()->getY()));

    // Filling arrays with 0 for the molecular and isotope spectra
    for (i = 0; i < par->getMol()->count(); i++)
    {
        horizSpectra->getMolAt(i)->getSpec()->setMaxValue(Auxiliary::maximum(horizSpectra->getMolAt(i)->getSpec()->getY()));
        // horizSpectra->getMolAt(i)->getSpec()->setMinValue(Auxiliary::minimum(horizSpectra->getMolAt(i)->getSpec()->getY()));
    }
}

// Method to calculate spectra for a horizontal path
void Lbl::calculateHorizPath()
{
    // Initizlizing the wavenumber grid
    wavelength->initWavenumberGrid();

    // Initializing the spectral arrays
    initHorizSpectra();

    // Running calculation for a horizontal path layer depending on the lineshape profiles used
    performLblCalculation();

    // Finding the maximum spectra
    findMinMaxValuesInSpectra();
}

// Method to calculate spectra for a slant path
void Lbl::calculateSlantPath()
{
    ;
}

// Method to calculate spectra for a single layer using the Lorentz profile
// This implementation is based in part on the HAPI implementation, the thoretical section of the freely available manual for HITRAN-PC was also consulted.
void Lbl::performLblCalculation()
{
    // If the cancel button has been pressed, then the calculation is canceled
    if (this->cancelCalculation == true)
    {
        postCalculation();
    }
    else
    {
        // Initial request to get the data (had to implement this through a signal slog because option 3 with download every time needs to time to download HITRAN data from the server)
        currRecNum = 0;
        hitran->getNextSubrange();
    }
    QApplication::processEvents();   // Updating the processed plot
}

// Slot is executed when the calculation is over
void Lbl::postCalculation()
{
    // Hiding the notification
    emit notify("", 0, false, true);
    QApplication::processEvents();

    //    // Forming the composite spectrum (this is done in the spectraresultsdataobject.cpp to enable combining of the spectra after the calculation)
    //    for (j = 0; j < par->getMol()->count(); j++)
    //    {
    //        // Summing spectra by wavelength
    //        for (i = 0; i < wavelength->getWns()->count(); i++)
    //        {
    //            optDepth = (*layerSpectra).getComp()->getYAt(i) + (*layerSpectra).getMolAt(j)->getSpec()->getYAt(i);
    //            (*layerSpectra).getComp()->setYAt(i, optDepth);
    //        }
    //    }

    // Finding the maximum spectra
    findMinMaxValuesInSpectra();

    // Emitting a signal to update the plotted spectra
    emit updateSpectra(true, true);
    QApplication::processEvents();

    // Emiting a signal stating that the calculation is complete
    emit calculationComplete();
}

// Slot to calculate a segment (for different lineshape profiles)
void Lbl::calcWnSegment()
{
    // Running calculation for a horizontal path layer depending on the lineshape profiles used
    switch (par->getLineshapeProfile())
    {
        case 0:
            calcDopplerWnSegment(); break;
        case 1:
            calcLorentzWnSegment(); break;
        case 2:
            calcVoigtWnSegment(); break;
        case 3:
            calcVoigtWnSegmentHighRes(); break;
        default:
            calcVoigtWnSegment(); break;
    }
}

// Slot to calculate on e Lorentz wavenumber segment
void Lbl::calcLorentzWnSegment()
{
    int i;  // Loop variable
    int j;  // Loop variable
    Spectrum *spec;   // Spectrum to contain the calculated result for a given record in the HITRAN database

    double optDepth;  // Optical depth value

    double factor;    // Volume concentration multiplication factor
    double gamma0 = 1;    // Gamma 0
    double shift0;    // Shift 0
    double wnWingF;   // wnWingF
    int boundIndexLower;  // Lower bounding index
    int boundIndexUpper;  // Upper bounding index

    // Partition functions
    double sigmaT;    // TIPS at the current temperature
    double sigmaTRef; // TIPS at the reference temperature
    double multiplicationFactor;
    double gamma2Squared;
    double wnWithShift;  // Wavenumber with the shift applied

    int molIndex;  // Molecule index
    int isoIndex;  // Isotope index

    double tempRatio;  // Temperature ratio

    // Line intensity calculation parameters
    double lineIntensity;
    double ch;
    double zn;
    double c = 1.4388028496642257;

    // Array with the read in HITRAN data
    QVector<HitranOneRec> *recs;

    // Ratio of the total pressure to the reference pressure and the same for the temperature
    double pressRatio;
    double multipyConst;

    // Value indicating the progress
    int totalNumRecs;  // Total number of records



    // Initializng an object to contain the spectral calculation results
    spec = new Spectrum();

    // Calculating the total number of records and the number of records after which the plot is to be updated
    totalNumRecs = hitran->getTotalNumRecsToProcess();

    // Reading the next set of indices
    if (hitran->getHitranFileUsage() == 2) currRecNum = 0;

    // Getting a pointer to the list of records read in
    recs = hitran->getHitranRecs();

    // Calculating the temperature and pressure ratio
    pressRatio = par->getHorizParams()->getTotalPress() / Const::totalPressRef;
    tempRatio = Const::tempRef / par->getHorizParams()->getTemp();

    // Calculating the multiplication constant
    multipyConst = 100 * 0.000001 * 296.0 * Const::Loch / Const::pi;

    // This was slowing dow the calculation considerably on the desktop, mobile doesn't seem to matter
    // Incrementing the sub-segment HITRAN record number
    QApplication::processEvents();
    emit notify("Found " + QString::number(totalNumRecs) + " HITRAN records.\nRunning the line-by-line calculation.\nProcessing segment " + QString::number(currRecNum + 1) + "-" + QString::number(recs->count() + currRecNum + 1) + ".", 0, true, true);
    QApplication::processEvents();

    // Running calculations for all available spectral lines
    for (i = 0; i < recs->count(); i++)
    {
        // Incrementing the current record number
        currRecNum++;

        // This may need to be optimized for speed by removing the methods, sometimes inline code works faster
        // such optimization should only be performed after everything is working in its initial implementation

        // Saving the molecule and isotope indices
        molIndex = ((*recs)[i]).getMolNum() - 1;
        isoIndex = ((*recs)[i]).getIsoNum() - 1;
        if ( ((*recs)[i]).getIsoNum() == 0 ) { isoIndex = 10; }  // Special case of the CO2 molecule

        // If the abundance is unavailable, then skipping the current calculation
        if (par->getMolAt(molIndex)->getIsoAt(isoIndex)->getTerrAbund() == 0) { continue; }

        // Precalculation of volume concentration
        // Depending on the HITRAN database units taking into account the multiplication factor
        if (par->getHitranUnits() == true)
        {
            // Multiplication factor is equal to 1.0
            factor = 1.0;
        }
        else
        {
            // Multiplication factor calculated from the pressure and temperature
            factor = par->getHorizParams()->getTotalPress() / (9.869233e-7 * Const::cBolts * par->getHorizParams()->getTemp());   // CGS
        }

        // Pressure shift, taking parameters from the HITRAN database if the line shift is used
        if (par->getUseLineShift()) { shift0 = ((*recs)[i]).getAirPShift() * pressRatio; }
        else { shift0 = 0; }

        // Getting all environment dependences from Voigt parameters, line intensity, used to be a separate procedure environmentDependencyIntensity
        ch = exp(-c * ((*recs)[i]).getLowE() / par->getHorizParams()->getTemp()) * (1 - exp(-c * ((*recs)[i]).getWn() / par->getHorizParams()->getTemp()));
        zn = exp(-c * ((*recs)[i]).getLowE() / Const::tempRef) * (1 - exp(-c * ((*recs)[i]).getWn() / Const::tempRef));

        // Partition functions for T and Tref, need to be optimized
        sigmaTRef = par->getMolAt(molIndex)->getIsoAt(isoIndex)->getQ296K();   // partitionFunction(MoleculeNumberDB, IsoNumberDB, Tref)
        sigmaT = par->getMolAt(molIndex)->getIsoAt(isoIndex)->getTipsAtIndex(round(par->getHorizParams()->getTemp()) - 70);   // partitionFunction(MoleculeNumberDB,IsoNumberDB,T)

        // Calculating the line intensity taking into account the TIPS values
        lineIntensity = ((*recs)[i]).getIntens() * (sigmaTRef / sigmaT) * (ch / zn);

        // Filtering by LineIntensity: compare it with IntencityThreshold. Need to apply wing narrowing instead of filtering, this would be more appropriate
        if (lineIntensity < par->getIntensThresh()) { continue; }  // Skip this line then, not included

        // Lorentz broadening coefficient, inlined method "environmentDependencyGamma0"
        // a^x -> exp(x * ln(a)) = a^x
        // Setting the broadening
        if (par->getLineBroadening() == 0)  // Air broadening only
        {
            gamma0 = ((*recs)[i]).getAirHw() * pressRatio * (1 - par->getHorizParams()->getPartPress(molIndex) * 0.000001) * pow((double) tempRatio, ((*recs)[i]).getNAir());
        }
        if (par->getLineBroadening() == 1)  // Self
        {
            gamma0 = ((*recs)[i]).getSelfHw() * pressRatio * par->getHorizParams()->getPartPress(molIndex) * 0.000001 * pow((double) tempRatio, ((*recs)[i]).getNAir());
        }
        if (par->getLineBroadening() == 2)  // Both air and self
        {
            gamma0 = pressRatio * ( (((*recs)[i]).getAirHw() * (1 - par->getHorizParams()->getPartPress(molIndex) * 0.000001) + ((*recs)[i]).getSelfHw() * par->getHorizParams()->getPartPress(molIndex) * 0.000001) * pow((double) tempRatio, ((*recs)[i]).getNAir()) );
        }

        // Get final wing of the line according to Gamma0, OmegaWingHW and OmegaWing, the maximum of the 2 values is used
        if ((par->getWingContribAbs() > (par->getWingContribSatHw() * gamma0))) { wnWingF = par->getWingContribAbs(); }
        else { wnWingF = par->getWingContribSatHw() * gamma0; }

        // Determining the lower and upper boundary indexes for a single spectral line based on the wing contribution for a single line
        wnWithShift = ((*recs)[i]).getWn() + shift0;
        // boundIndexCenter = wavelength->getWns()->count() * ((wnWithShift - wavelength->getStartWn()) / (wavelength->getEndWn() - wavelength->getStartWn()));
        boundIndexLower = wavelength->getWns()->count() * ((wnWithShift - wnWingF - wavelength->getStartWn()) / (wavelength->getEndWn() - wavelength->getStartWn()));
        // boundIndexUpper = 2 * boundIndexCenter - boundIndexLower;
        boundIndexUpper = wavelength->getWns()->count() * ((wnWithShift + wnWingF - wavelength->getStartWn()) / (wavelength->getEndWn() - wavelength->getStartWn()));
        spec->setStartIndex(boundIndexLower);
        spec->setEndIndex(boundIndexUpper);
        spec->getY()->resize(boundIndexUpper - boundIndexLower + 1);

        // Calculating Lorentz lineshape
        //    def PROFILE_LORENTZ(sg0,Gam0,sg):
        //
        //    # Lorentz profile.
        //    # Input parameters:
        //    # sg0: Unperturbed line position in cm-1 (Input).
        //    # Gam0: Speed-averaged line-width in cm-1 (Input).
        //    # sg: Current WaveNumber of the Computation in cm-1 (Input).
        //
        //    return Gam0/(pi*(Gam0**2+(sg-sg0)**2))

        // Optical depth determination:
        // OptDepth = xSect * Const::Loch * (296 / Temp) * ppmv * pathLength

        // sg0 - unperturbed line position
        //  sg - wavenumber grid (current wavenumbers to run calculations for)
        multiplicationFactor = multipyConst * gamma0 * factor * lineIntensity * par->getMolAt(molIndex)->getIsoAt(isoIndex)->getAbund() / (par->getMolAt(molIndex)->getIsoAt(isoIndex)->getTerrAbund());
        multiplicationFactor = (multiplicationFactor / par->getHorizParams()->getTemp()) * par->getHorizParams()->getPartPress(molIndex) * par->getHorizParams()->getPathLength() * pressRatio;  // 100 is to convert to centimeters and 0.000001 is to convert from parts per million (ppmv)
        gamma2Squared = gamma0 * gamma0;

        // Later take into account symmetry in the lineshape profile to make the calculations 2 times faster (only calculate half of the spectral line profile)

        // Limiting the calculated region to the current spectral range
        if (spec->getStartIndex() < 0) { spec->setStartIndex(0); }
        if (spec->getEndIndex() > (wavelength->getWns()->count() - 1)) { spec->setEndIndex(wavelength->getWns()->count() - 1); }
        for (j = spec->getStartIndex(); j <= spec->getEndIndex(); j++)
        {
            // Calculating the value for the current wavenumber position
            optDepth =  multiplicationFactor / (gamma2Squared + (wavelength->getWnsAt(j) - wnWithShift) * (wavelength->getWnsAt(j) - wnWithShift));
            // spec->setYAt(j - spec->getStartIndex(), optDepth);

            // Appending the spectrum to the total molecules spectrum
            optDepth = optDepth + horizSpectra->getMolAt(molIndex)->getSpec()->getYAt(j);
            horizSpectra->getMolAt(molIndex)->getSpec()->setYAt(j, optDepth);

            // Later on will probably change this to make it a separate isotopes addition
            // Maybe it makes sense to later include the sides in the caluclated single spectral lines
        }

        // Alternative implementation where spectral lines at the ends of the spectral regions are used completely
        //        if (oneLineSuccess)
        //        {
        //            if (spec->getStartIndex() < 0) { startIndex = 0; }
        //            if (spec->getEndIndex() > (wavelength->getWns()->count() - 1)) { endIndex = wavelength->getWns()->count() - 1; }
        //            // Appending the spectrum to the total spectrum for a given molecule
        //            // (the composite is calculated later on by adding all molecular spectra together)
        //            for (j = startIndex; j <= endIndex; j++)
        //            {
        //                molIndex = (*recs)[i].getMolNum() - 1;
        //                value = (*layerSpectra).getMolAt(molIndex)->getSpec()->getYAt(j) + spec->getYAt(j - spec->getStartIndex());
        //                (*layerSpectra).getMolAt(molIndex)->getSpec()->setYAt(j, value);
        //            }
        //            // Separate isotopes and separate lines spectra are not in the picture right now but will be added later on
        //        }
    }

    // Updating the plot
    findMinMaxValuesInSpectra();     // Finding the maximum spectra
    emit updateSpectra(true, true);  // Emitting a signal to update the plotted spectra
    QApplication::processEvents();   // Updating the processed plot

    // If the cancel button has been pressed, then the calculation is canceled
    if (this->cancelCalculation == true)
    {
        postCalculation();
    }
    else
    {
        // Requesting the next segment
        hitran->getNextSubrange();
    }
    QApplication::processEvents();   // Updating the processed plot
}

// Slot to calculate on e Lorentz wavenumber segment
void Lbl::calcDopplerWnSegment()
{
    int i;  // Loop variable
    int j;  // Loop variable
    Spectrum *spec;   // Spectrum to contain the calculated result for a given record in the HITRAN database

    double optDepth;  // Optical depth value

    double factor;    // Volume concentration multiplication factor
    double gammaD;    // GammaD
    double shift0;    // Shift 0
    double wnWingF;   // wnWingF
    int boundIndexLower;  // Lower bounding index
    int boundIndexUpper;  // Upper bounding index

    // Partition functions
    double sigmaT;    // TIPS at the current temperature
    double sigmaTRef; // TIPS at the reference temperature
    double multiplicationFactor;
    double wnWithShift;  // Wavenumber with the shift applied

    int molIndex;  // Molecule index
    int isoIndex;  // Isotope index

    // Line intensity calculation parameters
    double lineIntensity;
    double ch;
    double zn;
    double c = 1.4388028496642257;
    double fSqrtMass;

    // Array with the read in HITRAN data
    QVector<HitranOneRec> *recs;

    // Ratio of the total pressure to the reference pressure and the same for the temperature
    double pressRatio;
    double multipyConst;

    // Value indicating the progress
    int totalNumRecs;  // Total number of records
    QMessageBox msgBox;


    // Initializng an object to contain the spectral calculation results
    spec = new Spectrum();

    // Calculating the total number of records and the number of records after which the plot is to be updated
    totalNumRecs = hitran->getTotalNumRecsToProcess();

    // Reading the next set of indices
    if (hitran->getHitranFileUsage() == 2) currRecNum = 0;

    // Getting a pointer to the list of records read in
    recs = hitran->getHitranRecs();

    // Calculating the temperature and pressure ratio
    pressRatio = par->getHorizParams()->getTotalPress() / Const::totalPressRef;

    // Calculating the multiplication constant
    multipyConst = 100 * 0.000001 * 296.0 * Const::Loch;

    // This was slowing dow the calculation considerably on the desktop, mobile doesn't seem to matter
    // Incrementing the sub-segment HITRAN record number
    QApplication::processEvents();
    emit notify("Found " + QString::number(totalNumRecs) + " HITRAN records.\nRunning the line-by-line calculation.\nProcessing segment " + QString::number(currRecNum + 1) + "-" + QString::number(recs->count() + currRecNum + 1) + ".", 0, true, true);
    QApplication::processEvents();

    // Running calculations for all available spectral lines
    for (i = 0; i < recs->count(); i++)
    {
        // Incrementing the current record number
        currRecNum++;

        // This may need to be optimized for speed by removing the methods, sometimes inline code works faster
        // such optimization should only be performed after everything is working in its initial implementation

        // Saving the molecule and isotope indices
        molIndex = ((*recs)[i]).getMolNum() - 1;
        isoIndex = ((*recs)[i]).getIsoNum() - 1;
        if ( ((*recs)[i]).getIsoNum() == 0 ) { isoIndex = 10; }  // Special case of the CO2 molecule

        // If the abundance is unavailable, then skipping the current calculation
        if (par->getMolAt(molIndex)->getIsoAt(isoIndex)->getTerrAbund() == 0) { continue; }

        // Precalculation of volume concentration
        // Depending on the HITRAN database units taking into account the multiplication factor
        if (par->getHitranUnits() == true)
        {
            // Multiplication factor is equal to 1.0
            factor = 1.0;
        }
        else
        {
            // Multiplication factor calculated from the pressure and temperature
            factor = par->getHorizParams()->getTotalPress() / (9.869233e-7 * Const::cBolts * par->getHorizParams()->getTemp());   // CGS
        }

        // Pressure shift, taking parameters from the HITRAN database if the line shift is used
        if (par->getUseLineShift()) { shift0 = ((*recs)[i]).getAirPShift() * pressRatio; }
        else { shift0 = 0; }

        // Partition functions for T and Tref, need to be optimized
        sigmaT = par->getMolAt(molIndex)->getIsoAt(isoIndex)->getTips()->at(round(par->getHorizParams()->getTemp()) - 70);   // partitionFunction(MoleculeNumberDB,IsoNumberDB,T)
        sigmaTRef = par->getMolAt(molIndex)->getIsoAt(isoIndex)->getQ296K();   // partitionFunction(MoleculeNumberDB, IsoNumberDB, Tref)

        // Getting all environment dependences from Voigt parameters, line intensity, used to be a separate procedure environmentDependencyIntensity
        ch = exp(-c * ((*recs)[i]).getLowE() / par->getHorizParams()->getTemp()) * (1 - exp(-c * ((*recs)[i]).getWn() / par->getHorizParams()->getTemp()));
        zn = exp(-c * ((*recs)[i]).getLowE() / Const::tempRef) * (1 - exp(-c * ((*recs)[i]).getWn() / Const::tempRef));
        lineIntensity = ((*recs)[i]).getIntens() * (sigmaTRef / sigmaT) * (ch / zn);

        // Filtering by LineIntensity: compare it with IntencityThreshold. Need to apply wing narrowing instead of filtering, this would be more appropriate
        if (lineIntensity < par->getIntensThresh()) { continue; }  // Skip this line then, not included

        // GammaD calculation
        fSqrtMass = sqrt(par->getMolAt(molIndex)->getIsoAt(isoIndex)->getMolarMass());
        gammaD = (Const::cSqrt2Ln2 / Const::cc_) * sqrt(Const::cBolts_ / Const::cMassMol) * sqrt(par->getHorizParams()->getTemp()) * ((*recs)[i]).getWn() / fSqrtMass;

        // Get final wing of the line according to Gamma0, OmegaWingHW and OmegaWing, the maximum of the 2 values is used
        if ((par->getWingContribAbs() > (par->getWingContribSatHw() * gammaD))) { wnWingF = par->getWingContribAbs(); }
        else { wnWingF = par->getWingContribSatHw() * gammaD; }

        // Determining the lower and upper boundary indexes for a single spectral line based on the wing contribution for a single line
        wnWithShift = ((*recs)[i]).getWn() + shift0;
        // boundIndexCenter = wavelength->getWns()->count() * ((wnWithShift - wavelength->getStartWn()) / (wavelength->getEndWn() - wavelength->getStartWn()));
        boundIndexLower = wavelength->getWns()->count() * ((wnWithShift - wnWingF - wavelength->getStartWn()) / (wavelength->getEndWn() - wavelength->getStartWn()));
        // boundIndexUpper = 2 * boundIndexCenter - boundIndexLower;
        boundIndexUpper = wavelength->getWns()->count() * ((wnWithShift + wnWingF - wavelength->getStartWn()) / (wavelength->getEndWn() - wavelength->getStartWn()));
        spec->setStartIndex(boundIndexLower);
        spec->setEndIndex(boundIndexUpper);
        spec->getY()->resize(boundIndexUpper - boundIndexLower + 1);

        // Doppler profile.
        // Input parameters:
        // sg0: Unperturbed line position in cm-1 (Input).
        // GamD: Doppler HWHM in cm-1 (Input)
        // sg: Current WaveNumber of the Computation in cm-1 (Input).
        //
        // return cSqrtLn2divSqrtPi * exp(-cLn2*((sg-sg0) / GamD)**2) / GamD

        // Optical depth determination:
        // OptDepth = xSect * Const::Loch * (296 / Temp) * ppmv * pathLength

        // sg0 - unperturbed line position
        //  sg - wavenumber grid (current wavenumbers to run calculations for)
        multiplicationFactor = (multipyConst / gammaD) * factor * lineIntensity * par->getMolAt(molIndex)->getIsoAt(isoIndex)->getAbund() / (par->getMolAt(molIndex)->getIsoAt(isoIndex)->getTerrAbund());
        multiplicationFactor = (multiplicationFactor / par->getHorizParams()->getTemp()) * par->getHorizParams()->getPartPress(molIndex) * par->getHorizParams()->getPathLength() * pressRatio;  // 100 is to convert to centimeters and 0.000001 is to convert from parts per million (ppmv), embedded into the multiplyConst

        // Later take into account symmetry in the lineshape profile in the expectation to make the calculations 2 times faster (only calculate half of the spectral line profile)

        // Limiting the calculated region to the current spectral range
        if (spec->getStartIndex() < 0) { spec->setStartIndex(0); }
        if (spec->getEndIndex() > (wavelength->getWns()->count() - 1)) { spec->setEndIndex(wavelength->getWns()->count() - 1); }
        for (j = spec->getStartIndex(); j <= spec->getEndIndex(); j++)
        {
            // Calculating the value for the current wavenumber position
            // cSqrtLn2divSqrtPi * exp(-cLn2*((sg-sg0) / GamD)**2)
            optDepth = multiplicationFactor * Const::cSqrtLn2divSqrtPi * exp(-Const::cLn2 * ((wavelength->getWnsAt(j) - wnWithShift) / gammaD) * ((wavelength->getWnsAt(j) - wnWithShift) / gammaD));
            // spec->setYAt(j - spec->getStartIndex(), optDepth);

            // Appending the spectrum to the total molecules spectrum
            optDepth = optDepth + horizSpectra->getMolAt(molIndex)->getSpec()->getYAt(j);
            horizSpectra->getMolAt(molIndex)->getSpec()->setYAt(j, optDepth);

            // Later on will probably change this to make it a separate isotopes addition
            // Maybe it makes sense to later include the sides in the caluclated single spectral lines
        }

        // Alternative implementation where spectral lines at the ends of the spectral regions are used completely
        //        if (oneLineSuccess)
        //        {
        //            if (spec->getStartIndex() < 0) { startIndex = 0; }
        //            if (spec->getEndIndex() > (wavelength->getWns()->count() - 1)) { endIndex = wavelength->getWns()->count() - 1; }
        //            // Appending the spectrum to the total spectrum for a given molecule
        //            // (the composite is calculated later on by adding all molecular spectra together)
        //            for (j = startIndex; j <= endIndex; j++)
        //            {
        //                molIndex = (*recs)[i].getMolNum() - 1;
        //                value = (*layerSpectra).getMolAt(molIndex)->getSpec()->getYAt(j) + spec->getYAt(j - spec->getStartIndex());
        //                (*layerSpectra).getMolAt(molIndex)->getSpec()->setYAt(j, value);
        //            }
        //            // Separate isotopes and separate lines spectra are not in the picture right now but will be added later on
        //        }
    }

    // Updating the plot
    findMinMaxValuesInSpectra();     // Finding the maximum spectra
    emit updateSpectra(true, true);  // Emitting a signal to update the plotted spectra
    QApplication::processEvents();   // Updating the processed plot

    // If the cancel button has been pressed, then the calculation is canceled
    if (this->cancelCalculation == true)
    {
        postCalculation();
    }
    else
    {
        // Requesting the next segment
        hitran->getNextSubrange();
    }
    QApplication::processEvents();   // Updating the processed plot
}

// The following routine relies on the Voigt profile calculation algorithm and code described in the following publication:
// Abrarov, S., & Quine, B. (2015). A Rational Approximation for Efficient Computation of the Voigt Function
// in Quantitative Spectroscopy. Journal Of Mathematics Research, 7(2), p163. doi:http://dx.doi.org/10.5539/jmr.v7n2
// Slot to calculate one Voigt wavenumber segment (Low resolution implementation)
// Also: https://www.princeton.edu/cefrc/Files/2013%20Lecture%20Notes/Hanson/pLecture6.pdf
void Lbl::calcVoigtWnSegment()
{
    int i;  // Loop variable
    int j;  // Loop variable
    int t;  // Loop variable
    Spectrum *spec;    // Spectrum to contain the calculated result for a given record in the HITRAN database

    double optDepth;   // Optical depth value
    double value;      //

    double factor;     // Volume concentration multiplication factor
    double gammaD;     // Gamma D
    double gamma0 = 1; // Gamma 0
    double shift0;     // Shift 0
    double wnWingF = 0;   // wnWingF
    int boundIndexLower;  // Lower bounding index
    int boundIndexUpper;  // Upper bounding index

    // Partition functions
    double sigmaT;    // TIPS at the current temperature
    double sigmaTRef; // TIPS at the reference temperature
    double multiplicationFactor;
    double fSqrtMass;
    double wnWithShift;  // Wavenumber with the shift applied

    int molIndex;  // Molecule index
    int isoIndex;  // Isotope index

    double tempRatio;  // Temperature ratio

    // Line intensity calculation parameters
    double lineIntensity;
    double ch;
    double zn;
    double c = 1.4388028496642257;

    double multiplierVoigt;  // Multiplication term for the Voigt profile

    // Array with the read in HITRAN data
    QVector<HitranOneRec> *recs;

    // Auxiliaty variables for the calculation of the Voigt profile
    double vf;  // Value of the voigt function
    double xTerm;
    double x;
    double y;
    double varSigma = 2.75;
    double arr1;
    double arr2;
    double arr3;
    double ySquared;

    // Ratio of the total pressure to the reference pressure and the same for the temperature
    double pressRatio;
    double multipyConst;

    // Value indicating the progress
    int totalNumRecs;  // Total number of records
    QMessageBox msgBox;


    // Initializng an object to contain the spectral calculation results
    spec = new Spectrum();

    // Calculating the total number of records and the number of records after which the plot is to be updated
    totalNumRecs = hitran->getTotalNumRecsToProcess();

    // Reading the next set of indices
    if (hitran->getHitranFileUsage() == 2) currRecNum = 0;

    // Getting a pointer to the list of records read in
    recs = hitran->getHitranRecs();

    // Calculating the temperature and pressure ratio
    pressRatio = par->getHorizParams()->getTotalPress() / Const::totalPressRef;
    tempRatio = Const::tempRef / par->getHorizParams()->getTemp();

    // Calculating the multiplication constant
    multipyConst = 100 * 0.000001 * 296.0 * Const::Loch;  // Path is converted to cm, 0.000001 is to convert from parts per million

    // This was slowing dow the calculation considerably on the desktop, mobile doesn't seem to matter
    // Incrementing the sub-segment HITRAN record number
    QApplication::processEvents();
    emit notify("Found " + QString::number(totalNumRecs) + " HITRAN records.\nRunning the line-by-line calculation.\nProcessing segment " + QString::number(currRecNum + 1) + "-" + QString::number(recs->count() + currRecNum + 1) + ".", 0, true, true);
    QApplication::processEvents();

    // Running calculations for all available spectral lines
    for (i = 0; i < recs->count(); i++)
    {
        // Incrementing the current record number
        currRecNum++;

        // This may need to be optimized for speed by removing the methods, sometimes inline code works faster
        // such optimization should only be performed after everything is working in its initial implementation

        // Saving the molecule and isotope indices
        molIndex = ((*recs)[i]).getMolNum() - 1;
        isoIndex = ((*recs)[i]).getIsoNum() - 1;
        if ( ((*recs)[i]).getIsoNum() == 0 ) { isoIndex = 10; }  // Special case of the CO2 molecule

        // If the abundance is unavailable, then skipping the current calculation
        if (par->getMolAt(molIndex)->getIsoAt(isoIndex)->getTerrAbund() == 0) { continue; }

        // Precalculation of volume concentration
        // Depending on the HITRAN database units taking into account the multiplication factor
        if (par->getHitranUnits() == true)
        {
            // Multiplication factor is equal to 1.0
            factor = 1.0;
        }
        else
        {
            // Multiplication factor calculated from the pressure and temperature
            factor = par->getHorizParams()->getTotalPress() / (9.869233e-7 * Const::cBolts * par->getHorizParams()->getTemp());   // CGS
        }

        // Pressure shift, taking parameters from the HITRAN database if the line shift is used
        if (par->getUseLineShift()) { shift0 = ((*recs)[i]).getAirPShift() * pressRatio; }
        else { shift0 = 0; }

        // Partition functions for T and Tref, need to be optimized
        sigmaT = par->getMolAt(molIndex)->getIsoAt(isoIndex)->getTips()->at(round(par->getHorizParams()->getTemp()) - 70);   // partitionFunction(MoleculeNumberDB,IsoNumberDB,T)
        sigmaTRef = par->getMolAt(molIndex)->getIsoAt(isoIndex)->getQ296K();   // partitionFunction(MoleculeNumberDB, IsoNumberDB, Tref)

        // Getting all environment dependences from Voigt parameters, line intensity, used to be a separate procedure environmentDependencyIntensity
        ch = exp(-c * ((*recs)[i]).getLowE() / par->getHorizParams()->getTemp()) * (1 - exp(-c * ((*recs)[i]).getWn() / par->getHorizParams()->getTemp()));
        zn = exp(-c * ((*recs)[i]).getLowE() / Const::tempRef) * (1 - exp(-c * ((*recs)[i]).getWn() / Const::tempRef));
        lineIntensity = ((*recs)[i]).getIntens() * (sigmaTRef / sigmaT) * (ch / zn);

        // Filtering by LineIntensity: compare it with IntensityThreshold. Need to apply wing narrowing instead of filtering, this would be more appropriate
        if (lineIntensity < par->getIntensThresh()) { continue; }  // Skip this line then, not included

        // GammaD calculation
        fSqrtMass = sqrt(par->getMolAt(molIndex)->getIsoAt(isoIndex)->getMolarMass());
        gammaD = (Const::cSqrt2Ln2 / Const::cc_) * sqrt(Const::cBolts_ / Const::cMassMol) * sqrt(par->getHorizParams()->getTemp()) * ((*recs)[i]).getWn() / fSqrtMass;

        // Lorentz broadening coefficient, inlined method "environmentDependencyGamma0"
        // a^x -> exp(x * ln(a)) = a^x
        // Setting the broadening
        if (par->getLineBroadening() == 0)  // Air broadening only
        {
            gamma0 = ((*recs)[i]).getAirHw() * pressRatio * (1 - par->getHorizParams()->getPartPress(molIndex) * 0.000001) * pow((double) tempRatio, ((*recs)[i]).getNAir());
        }
        if (par->getLineBroadening() == 1)  // Self
        {
            gamma0 = ((*recs)[i]).getSelfHw() * pressRatio * par->getHorizParams()->getPartPress(molIndex) * 0.000001 * pow((double) tempRatio, ((*recs)[i]).getNAir());
        }
        if (par->getLineBroadening() == 2)  // Both air and self
        {
            gamma0 = pressRatio * ( (((*recs)[i]).getAirHw() * (1 - par->getHorizParams()->getPartPress(molIndex) * 0.000001) + ((*recs)[i]).getSelfHw() * par->getHorizParams()->getPartPress(molIndex) * 0.000001) * pow((double) tempRatio, ((*recs)[i]).getNAir()) );
        }

        // Get final wing of the line according to Gamma0, OmegaWingHW and OmegaWing, the maximum of the 2 values is used
        // OmegaWingF = max(OmegaWing, OmegaWingHW * Gamma0, OmegaWingHW * GammaD);
        if ((par->getWingContribAbs() > (par->getWingContribSatHw() * gamma0)) && (par->getWingContribAbs() > (par->getWingContribSatHw() * gammaD))) { wnWingF = par->getWingContribAbs();; }
        if (((par->getWingContribSatHw() * gamma0) > par->getWingContribAbs()) && (gamma0 > gammaD)) { wnWingF = (par->getWingContribSatHw() * gamma0); }
        if (((par->getWingContribSatHw() * gammaD) > par->getWingContribAbs()) && (gammaD > gamma0)) { wnWingF = (par->getWingContribSatHw() * gammaD); }

        // Determining the lower and upper boundary indexes for a single spectral line based on the wing contribution for a single line
        wnWithShift = ((*recs)[i]).getWn() + shift0;
        // boundIndexCenter = wavelength->getWns()->count() * ((wnWithShift - wavelength->getStartWn()) / (wavelength->getEndWn() - wavelength->getStartWn()));
        boundIndexLower = wavelength->getWns()->count() * ((wnWithShift - wnWingF - wavelength->getStartWn()) / (wavelength->getEndWn() - wavelength->getStartWn()));
        // boundIndexUpper = 2 * boundIndexCenter - boundIndexLower;
        boundIndexUpper = wavelength->getWns()->count() * ((wnWithShift + wnWingF - wavelength->getStartWn()) / (wavelength->getEndWn() - wavelength->getStartWn()));
        spec->setStartIndex(boundIndexLower);
        spec->setEndIndex(boundIndexUpper);
        spec->getY()->resize(boundIndexUpper - boundIndexLower + 1);

        // Auxiliary intermediate terms for the calculation for the Voigt profile
        xTerm = Const::sqrtLn2 / gammaD;  // 2 multiplication term is absent because Princeton is FWHM and here it is HWHM
        y = Const::sqrtLn2 * gamma0 / gammaD;
        y = fabs(y) + varSigma * 0.5;
        ySquared = y * y;

        // This term is according to the lecture from Princeton
        // https://www.princeton.edu/cefrc/Files/2013%20Lecture%20Notes/Hanson/pLecture6.pdf
        multiplierVoigt = Const::sqrtLn2 / (sqrt(Const::pi) * gammaD);  // 2 multiplication term is absent because Princeton is FWHM and here it is HWHM

        // Optical depth determination:
        // OptDepth = xSect * Const::Loch * (296 / Temp) * ppmv * pathLength

        // sg0 - unperturbed line position
        //  sg - wavenumber grid (current wavenumbers to run calculations for)
        multiplicationFactor = multipyConst * factor * lineIntensity * par->getMolAt(molIndex)->getIsoAt(isoIndex)->getAbund() / (par->getMolAt(molIndex)->getIsoAt(isoIndex)->getTerrAbund());
        multiplicationFactor = (multiplicationFactor / par->getHorizParams()->getTemp()) * par->getHorizParams()->getPartPress(molIndex) * par->getHorizParams()->getPathLength() * pressRatio;  // 100 is to convert to centimeters and 0.000001 is to convert from parts per million (ppmv)

        // Later take into account symmetry in the lineshape profile to make the calculations 2 times faster (only calculate half of the spectral line profile)
        // Limiting the calculated region to the current spectral range

        if (spec->getStartIndex() < 0) { spec->setStartIndex(0); }
        if (spec->getEndIndex() > (wavelength->getWns()->count() - 1)) { spec->setEndIndex(wavelength->getWns()->count() - 1); }
        for (j = spec->getStartIndex(); j < spec->getEndIndex(); j++)
        {
            // The following routine relies on the Voigt profile calculation algorithm and code described in the following publication:
            // Abrarov, S., & Quine, B. (2015). A Rational Approximation for Efficient Computation of the Voigt Function
            // in Quantitative Spectroscopy. Journal Of Mathematics Research, 7(2), p163. doi:http://dx.doi.org/10.5539/jmr.v7n2
            // Slot to calculate one Voigt wavenumber segment (Low resolution implementation)
            // Preparing the terms
            x = xTerm * (wavelength->getWnsAt(j) - wnWithShift);
            arr1 = ySquared - x * x;
            arr2 = x * x + ySquared;
            arr3 = arr2 * arr2;
            // Calculating the Voigt profile value
            vf = 0;
            for (t = 0; t < coeff1_12.count(); t++)
            {
                vf = vf + (coeff1_12[t] * (coeff2_12[t] + arr1) + coeff3_12[t] * y * (coeff2_12[t] + arr2)) / (coeff2_12[t] * coeff2_12[t] + 2 * coeff2_12[t] * arr1 + arr3) ;
            }
            optDepth = multiplicationFactor * multiplierVoigt * vf;
            value = optDepth + horizSpectra->getMolAt(molIndex)->getSpec()->getYAt(j);
            horizSpectra->getMolAt(molIndex)->getSpec()->setYAt(j, value);

            // Calculating the value for the current wavenumber position
            // spec->setYAt(j - spec->getStartIndex(), optDepth);

            // Later on will probably change this to make it a separate isotopes addition
            // Maybe it makes sense to later include the sides in the caluclated single spectral lines
        }

        // Alternative implementation where spectral lines at the ends of the spectral regions are used completely
        //        if (oneLineSuccess)
        //        {
        //            if (spec->getStartIndex() < 0) { startIndex = 0; }
        //            if (spec->getEndIndex() > (wavelength->getWns()->count() - 1)) { endIndex = wavelength->getWns()->count() - 1; }
        //            // Appending the spectrum to the total spectrum for a given molecule
        //            // (the composite is calculated later on by adding all molecular spectra together)
        //            for (j = startIndex; j <= endIndex; j++)
        //            {
        //                molIndex = (*recs)[i].getMolNum() - 1;
        //                value = (*layerSpectra).getMolAt(molIndex)->getSpec()->getYAt(j) + spec->getYAt(j - spec->getStartIndex());
        //                (*layerSpectra).getMolAt(molIndex)->getSpec()->setYAt(j, value);
        //            }
        //            // Separate isotopes and separate lines spectra are not in the picture right now but will be added later on
        //        }

    }

    // Updating the plot
    findMinMaxValuesInSpectra();     // Finding the maximum spectra
    emit updateSpectra(true, true);  // Emitting a signal to update the plotted spectra
    QApplication::processEvents();   // Updating the processed plot

    // If the cancel button has been pressed, then the calculation is canceled
    if (this->cancelCalculation == true)
    {
        postCalculation();
    }
    else
    {
        // Requesting the next segment
        hitran->getNextSubrange();
    }
    QApplication::processEvents();   // Updating the processed plot
}

// The following routine relies on the Voigt profile calculation algorithm and code described in the following publication:
// Abrarov, S., & Quine, B. (2015). A Rational Approximation for Efficient Computation of the Voigt Function
// in Quantitative Spectroscopy. Journal Of Mathematics Research, 7(2), p163. doi:http://dx.doi.org/10.5539/jmr.v7n2
// This High resolution implementration relies on a 16-coefficient implementation
void Lbl::calcVoigtWnSegmentHighRes()
{
    int i;  // Loop variable
    int j;  // Loop variable
    int t;  // Loop variable
    Spectrum *spec;   // Spectrum to contain the calculated result for a given record in the HITRAN database
    double optDepth;  // Optical depth value

    double factor;     // Volume concentration multiplication factor
    double gammaD;     // Gamma D
    double gamma0 = 1; // Gamma 0
    double shift0;     // Shift 0
    double wnWingF = 0;   // wnWingF
    int boundIndexLower;  // Lower bounding index
    int boundIndexUpper;  // Upper bounding index

    // Partition functions
    double sigmaT;    // TIPS at the current temperature
    double sigmaTRef; // TIPS at the reference temperature
    double multiplicationFactor;
    double fSqrtMass;
    double wnWithShift;  // Wavenumber with the shift applied

    int molIndex;  // Molecule index
    int isoIndex;  // Isotope index

    double tempRatio;  // Temperature ratio

    // Line intensity calculation parameters
    double lineIntensity;
    double ch;
    double zn;
    double c = 1.4388028496642257;

    double multiplierVoigt;  // Multiplication term for the Voigt profile

    // Array with the read in HITRAN data
    QVector<HitranOneRec> *recs;

    // Auxiliaty variables for the calculation of the Voigt profile
    double vf;  // Value of the voigt function
    double xTerm;
    double x;
    double y;
    double varSigma = 2.75;
    double arr1;
    double arr2;
    double arr3;
    double ySquared;

    // Ratio of the total pressure to the reference pressure and the same for the temperature
    double pressRatio;
    double multipyConst;

    // Value indicating the progress
    int totalNumRecs;  // Total number of records
    QMessageBox msgBox;


    // Initializng an object to contain the spectral calculation results
    spec = new Spectrum();

    // Calculating the total number of records and the number of records after which the plot is to be updated
    totalNumRecs = hitran->getTotalNumRecsToProcess();

    // Reading the next set of indices
    if (hitran->getHitranFileUsage() == 2) currRecNum = 0;

    // Getting a pointer to the list of records read in
    recs = hitran->getHitranRecs();

    // Calculating the temperature and pressure ratio
    pressRatio = par->getHorizParams()->getTotalPress() / Const::totalPressRef;
    tempRatio = Const::tempRef / par->getHorizParams()->getTemp();

    // Calculating the multiplication constant
    multipyConst = 100 * 0.000001 * 296.0 * Const::Loch;

    // This was slowing dow the calculation considerably on the desktop, mobile doesn't seem to matter
    // Incrementing the sub-segment HITRAN record number
    QApplication::processEvents();
    emit notify("Found " + QString::number(totalNumRecs) + " HITRAN records.\nRunning the line-by-line calculation.\nProcessing segment " + QString::number(currRecNum + 1) + "-" + QString::number(recs->count() + currRecNum + 1) + ".", 0, true, true);
    QApplication::processEvents();

    // Running calculations for all available spectral lines
    for (i = 0; i < recs->count(); i++)
    {
        // Incrementing the current record number
        currRecNum++;

        // This may need to be optimized for speed by removing the methods, sometimes inline code works faster
        // such optimization should only be performed after everything is working in its initial implementation

        // Saving the molecule and isotope indices
        molIndex = ((*recs)[i]).getMolNum() - 1;
        isoIndex = ((*recs)[i]).getIsoNum() - 1;
        if ( ((*recs)[i]).getIsoNum() == 0 ) { isoIndex = 10; }  // Special case of the CO2 molecule

        // If the abundance is unavailable, then skipping the current calculation
        if (par->getMolAt(molIndex)->getIsoAt(isoIndex)->getTerrAbund() == 0) { continue; }

        // Precalculation of volume concentration
        // Depending on the HITRAN database units taking into account the multiplication factor
        if (par->getHitranUnits() == true)
        {
            // Multiplication factor is equal to 1.0
            factor = 1.0;
        }
        else
        {
            // Multiplication factor calculated from the pressure and temperature
            factor = par->getHorizParams()->getTotalPress() / (9.869233e-7 * Const::cBolts * par->getHorizParams()->getTemp());   // CGS
        }

        // Pressure shift, taking parameters from the HITRAN database if the line shift is used
        if (par->getUseLineShift()) { shift0 = ((*recs)[i]).getAirPShift() * pressRatio; }
        else { shift0 = 0; }

        // Partition functions for T and Tref, need to be optimized
        sigmaT = par->getMolAt(molIndex)->getIsoAt(isoIndex)->getTips()->at(round(par->getHorizParams()->getTemp()) - 70);   // partitionFunction(MoleculeNumberDB,IsoNumberDB,T)
        sigmaTRef = par->getMolAt(molIndex)->getIsoAt(isoIndex)->getQ296K();   // partitionFunction(MoleculeNumberDB, IsoNumberDB, Tref)

        // Getting all environment dependences from Voigt parameters, line intensity, used to be a separate procedure environmentDependencyIntensity
        ch = exp(-c * ((*recs)[i]).getLowE() / par->getHorizParams()->getTemp()) * (1 - exp(-c * ((*recs)[i]).getWn() / par->getHorizParams()->getTemp()));
        zn = exp(-c * ((*recs)[i]).getLowE() / Const::tempRef) * (1 - exp(-c * ((*recs)[i]).getWn() / Const::tempRef));
        lineIntensity = ((*recs)[i]).getIntens() * (sigmaTRef / sigmaT) * (ch / zn);

        // Filtering by LineIntensity: compare it with IntensityThreshold. Need to apply wing narrowing instead of filtering, this would be more appropriate
        if (lineIntensity < par->getIntensThresh()) { continue; }  // Skip this line then, not included

        // GammaD calculation
        fSqrtMass = sqrt(par->getMolAt(molIndex)->getIsoAt(isoIndex)->getMolarMass());
        gammaD = (Const::cSqrt2Ln2 / Const::cc_) * sqrt(Const::cBolts_ / Const::cMassMol) * sqrt(par->getHorizParams()->getTemp()) * ((*recs)[i]).getWn() / fSqrtMass;

        // Lorentz broadening coefficient, inlined method "environmentDependencyGamma0"
        // a^x -> exp(x * ln(a)) = a^x
        // Setting the broadening
        if (par->getLineBroadening() == 0)  // Air broadening only
        {
            gamma0 = ((*recs)[i]).getAirHw() * pressRatio * (1 - par->getHorizParams()->getPartPress(molIndex) * 0.000001) * pow((double) tempRatio, ((*recs)[i]).getNAir());
        }
        if (par->getLineBroadening() == 1)  // Self
        {
            gamma0 = ((*recs)[i]).getSelfHw() * pressRatio * par->getHorizParams()->getPartPress(molIndex) * 0.000001 * pow((double) tempRatio, ((*recs)[i]).getNAir());
        }
        if (par->getLineBroadening() == 2)  // Both air and self
        {
            gamma0 = pressRatio * ( (((*recs)[i]).getAirHw() * (1 - par->getHorizParams()->getPartPress(molIndex) * 0.000001) + ((*recs)[i]).getSelfHw() * par->getHorizParams()->getPartPress(molIndex) * 0.000001) * pow((double) tempRatio, ((*recs)[i]).getNAir()) );
        }

        // Get final wing of the line according to Gamma0, OmegaWingHW and OmegaWing, the maximum of the 2 values is used
        // OmegaWingF = max(OmegaWing, OmegaWingHW * Gamma0, OmegaWingHW * GammaD);
        if ((par->getWingContribAbs() > (par->getWingContribSatHw() * gamma0)) && (par->getWingContribAbs() > (par->getWingContribSatHw() * gammaD))) { wnWingF = par->getWingContribAbs();; }
        if (((par->getWingContribSatHw() * gamma0) > par->getWingContribAbs()) && (gamma0 > gammaD)) { wnWingF = (par->getWingContribSatHw() * gamma0); }
        if (((par->getWingContribSatHw() * gammaD) > par->getWingContribAbs()) && (gammaD > gamma0)) { wnWingF = (par->getWingContribSatHw() * gammaD); }

        // Determining the lower and upper boundary indexes for a single spectral line based on the wing contribution for a single line
        wnWithShift = ((*recs)[i]).getWn() + shift0;
        // boundIndexCenter = wavelength->getWns()->count() * ((wnWithShift - wavelength->getStartWn()) / (wavelength->getEndWn() - wavelength->getStartWn()));
        boundIndexLower = wavelength->getWns()->count() * ((wnWithShift - wnWingF - wavelength->getStartWn()) / (wavelength->getEndWn() - wavelength->getStartWn()));
        // boundIndexUpper = 2 * boundIndexCenter - boundIndexLower;
        boundIndexUpper = wavelength->getWns()->count() * ((wnWithShift + wnWingF - wavelength->getStartWn()) / (wavelength->getEndWn() - wavelength->getStartWn()));
        spec->setStartIndex(boundIndexLower);
        spec->setEndIndex(boundIndexUpper);
        spec->getY()->resize(boundIndexUpper - boundIndexLower + 1);

        // Auxiliary intermediate terms for the calculation for the Voigt profile
        xTerm = Const::sqrtLn2 / gammaD;  // 2 multiplication term is absent because Princeton is FWHM and here it is HWHM
        y = Const::sqrtLn2 * gamma0 / gammaD;
        y = fabs(y) + varSigma * 0.5;
        ySquared = y * y;

        // This term is according to the lecture from Princeton
        // https://www.princeton.edu/cefrc/Files/2013%20Lecture%20Notes/Hanson/pLecture6.pdf
        multiplierVoigt = Const::sqrtLn2 / (sqrt(Const::pi) * gammaD);  // 2 multiplication term is absent because Princeton is FWHM and here it is HWHM

        // Optical depth determination:
        // OptDepth = xSect * Const::Loch * (296 / Temp) * ppmv * pathLength

        // sg0 - unperturbed line position
        //  sg - wavenumber grid (current wavenumbers to run calculations for)
        multiplicationFactor = multipyConst * factor * lineIntensity * par->getMolAt(molIndex)->getIsoAt(isoIndex)->getAbund() / (par->getMolAt(molIndex)->getIsoAt(isoIndex)->getTerrAbund());
        multiplicationFactor = (multiplicationFactor / par->getHorizParams()->getTemp()) * par->getHorizParams()->getPartPress(molIndex) * par->getHorizParams()->getPathLength() * pressRatio;  // 100 is to convert to centimeters and 0.000001 is to convert from parts per million (ppmv)

        // Later take into account symmetry in the lineshape profile to make the calculations 2 times faster (only calculate half of the spectral line profile)
        // Limiting the calculated region to the current spectral range

        if (spec->getStartIndex() < 0) { spec->setStartIndex(0); }
        if (spec->getEndIndex() > (wavelength->getWns()->count() - 1)) { spec->setEndIndex(wavelength->getWns()->count() - 1); }
        for (j = spec->getStartIndex(); j <= spec->getEndIndex(); j++)
        {
            // The following routine relies on the Voigt profile calculation algorithm and code described in the following publication:
            // Abrarov, S., & Quine, B. (2015). A Rational Approximation for Efficient Computation of the Voigt Function
            // in Quantitative Spectroscopy. Journal Of Mathematics Research, 7(2), p163. doi:http://dx.doi.org/10.5539/jmr.v7n2
            // Slot to calculate one Voigt wavenumber segment (Low resolution implementation)
            // Preparing the terms
            x = xTerm * (wavelength->getWnsAt(j) - wnWithShift);
            arr1 = ySquared - x * x;
            arr2 = x * x + ySquared;
            arr3 = arr2 * arr2;
            // Calculating the Voigt profile value
            vf = 0;
            for (t = 0; t < coeff1_16.count(); t++)
            {
                vf = vf + (coeff1_16[t] * (coeff2_16[t] + arr1) + coeff3_16[t] * y * (coeff2_16[t] + arr2)) / (coeff2_16[t] * coeff2_16[t] + 2 * coeff2_16[t] * arr1 + arr3) ;
            }
            optDepth = multiplicationFactor * multiplierVoigt * vf;

            // Calculating the value for the current wavenumber position
            // spec->setYAt(j - spec->getStartIndex(), optDepth);

            // Appending the spectrum to the total molecules spectrum
            optDepth = optDepth + horizSpectra->getMolAt(molIndex)->getSpec()->getYAt(j);
            horizSpectra->getMolAt(molIndex)->getSpec()->setYAt(j, optDepth);

            // Later on will probably change this to make it a separate isotopes addition
            // Maybe it makes sense to later include the sides in the caluclated single spectral lines
        }

        // Alternative implementation where spectral lines at the ends of the spectral regions are used completely
        //        if (oneLineSuccess)
        //        {
        //            if (spec->getStartIndex() < 0) { startIndex = 0; }
        //            if (spec->getEndIndex() > (wavelength->getWns()->count() - 1)) { endIndex = wavelength->getWns()->count() - 1; }
        //            // Appending the spectrum to the total spectrum for a given molecule
        //            // (the composite is calculated later on by adding all molecular spectra together)
        //            for (j = startIndex; j <= endIndex; j++)
        //            {
        //                molIndex = (*recs)[i].getMolNum() - 1;
        //                value = (*layerSpectra).getMolAt(molIndex)->getSpec()->getYAt(j) + spec->getYAt(j - spec->getStartIndex());
        //                (*layerSpectra).getMolAt(molIndex)->getSpec()->setYAt(j, value);
        //            }
        //            // Separate isotopes and separate lines spectra are not in the picture right now but will be added later on
        //        }

    }

    // Updating the plot
    findMinMaxValuesInSpectra();     // Finding the maximum spectra
    emit updateSpectra(true, true);  // Emitting a signal to update the plotted spectra
    QApplication::processEvents();   // Updating the processed plot

    // If the cancel button has been pressed, then the calculation is canceled
    if (this->cancelCalculation == true)
    {
        postCalculation();
    }
    else
    {
        // Requesting the next segment
        hitran->getNextSubrange();
    }
    QApplication::processEvents();   // Updating the processed plot
}

// Method to calculate spectra for a single layer using the Hartmann-Tran profile
//void Lbl::calcOneLayerHartmannTran(QVector<HitranOneRec> *recs, OneLayerSpectra *layerSpectra)
//{
//    int i;  // Loop variable
//    int j;  // Loop variable
//    double value;  // Auxiliary variable
//    int molIndex;  // Molecule index
//    Spectrum *spec;   // Spectrum to contain the calculated result for a given record in the HITRAN database
//    // int isoIndex;  // Isotope index, to be used in the future implementations
//    bool oneLineSuccess;  // Variable indicates that one line calculation has been successful
//    QMessageBox msgBox;
//    int startIndex;   // Start wavenumber index for a given line
//    int endIndex;     // End wavenumber index for a given line


//    // Initializng an object to contain the spectral calculation results
//    spec = new Spectrum();

//    // Running calculations for all available spectral lines
//    for (i = 0; i < recs->count(); i++)
//    {
//        // This may need to be optimized for speed by removing the methods, sometimes inline code works faster
//        // such optimization should only be performed after everything is working in its initial implementation

//        // Method to calculate the Hartmann-Tran absorption coefficient
//        oneLineSuccess = calcSingleLineAbsorptionCoeffHartmannTran(&((*recs)[i]),
//                                                                    ((*recs)[i]).getMolNum(),
//                                                                    ((*recs)[i]).getIsoNum(),
//                                                                    par->getMolAt(((*recs)[i]).getMolNum() - 1)->getIsoAt(((*recs)[i]).getIsoNum() - 1)->getTerrAbund(),
//                                                                    par->getHorizParams()->getTotalPress(),
//                                                                    par->getHorizParams()->getTemp(),
//                                                                    wavelength->getStartWn(),
//                                                                    wavelength->getEndWn(),
//                                                                    wavelength->getWnResol(),
//                                                                    par->getWingContribAbs(),
//                                                                    par->getIntensThresh(),
//                                                                    par->getWingContribSatHw(),
//                                                                    par->getLineBroadening(),
//                                                                    par->getHitranUnits(),
//                                                                    par->getUseLineShift(),
//                                                                    spec);

//        if (oneLineSuccess)
//        {
//            if (spec->getStartIndex() < 0) { startIndex = 0; }
//                                      else { startIndex = spec->getStartIndex(); }
//            if (spec->getEndIndex() > (wavelength->getWns()->count() - 1)) { endIndex = wavelength->getWns()->count() - 1; }
//                                                                      else { endIndex = spec->getEndIndex(); }

//            // Appending the spectrum to the total spectrum for a given molecule
//            // (the composite is calculated later on by adding all molecular spectra together)
//            for (j = startIndex; j <= endIndex; j++)
//            {
//                molIndex = (*recs)[i].getMolNum() - 1;
//                value = (*layerSpectra).getMolAt(molIndex)->getSpec()->getYAt(j) + spec->getYAt(j - spec->getStartIndex());
//                (*layerSpectra).getMolAt(molIndex)->getSpec()->setYAt(j, value);
//            }
//            // Separate isotopes and separate lines spectra are not in the picture right now but will be added later on
//        }
//    }

//    // Forming the composite spectrum
//    for (i = 0; i < wavelength->getWns()->count(); i++)
//    {
//        for (j = 0; j < par->getMol()->count(); j++)
//        {
//            value = (*layerSpectra).getComp()->getYAt(i) + (*layerSpectra).getMolAt(j)->getSpec()->getYAt(i);
//            (*layerSpectra).getComp()->setYAt(i, value);
//        }
//    }
//}

// Method to calculate the Hartmann-Tran absorption coefficient, (HAPI converted)
// This method is converted from the Harvard Smithsonian's HAPI implementation
//bool Lbl::calcSingleLineAbsorptionCoeffHartmannTran(HitranOneRec *rec,
//                                                    int mol,
//                                                    int iso,
//                                                    double abund,
//                                                    double totalPress,
//                                                    double temp,
//                                                    double wnStart,
//                                                    double wnEnd,
//                                                    double wnStep,
//                                                    double wnWingAbs,
//                                                    double intensThresh,
//                                                    double wnWingStHw,
//                                                    double broadType,
//                                                    bool hitranUnits,
//                                                    bool useLineShift,
//                                                    Spectrum *spec)

//{

//    //          mol -> HITRAN molecule number
//    //          iso -> HITRAN isotope number
//    //        abund -> Isotopologue abundance
//    //        press -> Total pressure
//    //         temp -> Temperature of the gas
//    //      wnStart -> Starting wavenumber, cm-1
//    //        wnEnd -> Ending wavenumber, cm-1
//    //       wnStep -> Wavenumber step to consider, cm-1
//    //       wnWing -> Absolute wing contribution in cm-1 to take into account in calculations
//    // intensThresh -> Intensity threshold, minimum intensity of a line for it to be considered in calculations
//    //     wnWingHw -> Wing halfwidth in wavenumbers, cm-1
//    //       gammaL -> Broadening parameter ("gamma_air" or "gamma_self")
//    //  hitranUnits -> HITRAN units: True - cm^2 / molecule; False - cm-1;
//    //           wn -> Wavenumber grid, cm-1
//    //        xSect -> Calculated absorption coefficient grid

//    // DESCRIPTION (This description is taken directly from HAPI, HITRAN application programming interface):
//    // Calculate absorption coefficient using HT (Hartmann-Tran) profile.
//    // Absorption coefficient is calculated at arbitrary temperature and pressure.
//    // User can vary a wide range of parameters to control a process of calculation
//    // (such as OmegaRange, OmegaStep, OmegaWing, OmegaWingHW, IntensityThreshold).
//    // The choice of these parameters depends on properties of a particular linelist.
//    // Default values are a sort of guess which gives a decent precicion (on average)
//    // for a reasonable amount of cpu time. To increase calculation accuracy,
//    // user should use a trial and error method.

//    // Components: (optional parameter)
//    // List of tuples (M,I,D) to consider in cross section calculation.
//    // M - here is a molecule number,
//    // I - is an isotopologue number,
//    // D - is an abundance of the component.
//    // NOTE: If this input contains more than one tuple, then the output
//    // is an absorption coefficient for mixture of corresponding gases.
//    // NOTE2: If omitted, then all data from the source tables is involved.

//    // Environment: (optional parameter)
//    // Python dictionary containing value of pressure and temperature.
//    // P - value of pressure
//    // T - value of temperature
//    // The format is as follows: Environment = {'p':pval,'T':tval},
//    // where "pval" and "tval" are corresponding values in atm and K
//    // respectively.
//    // NOTE: Default value is {'p':1.0,'T':296.0}

//    // INPUT PARAMETERS:
//    //    Components: list of tuples [(M,I,D)], where
//    //    M - HITRAN molecule number,
//    //    I - HITRAN isotopologue number,
//    //    D - abundance (optional)
//    //    SourceTables: list of tables from which to calculate cross-section (optional)
//    //    partitionFunction: pointer to partition function (default is PYTIPS) (optional)
//    //    Environment: dictionary containing thermodynamic parameters.
//    //    'p' - pressure in atmospheres,
//    //    'T' - temperature in Kelvin
//    //    Default={'p':1.,'T':296.}
//    //    OmegaRange: wavenumber range to consider.
//    //    OmegaStep: wavenumber step to consider.
//    //    OmegaWing: absolute wing for calculating a lineshape (in cm-1)
//    //    IntensityThreshold: threshold for intensities
//    //    OmegaWingHW: relative wing for calculating a lineshape (in halfwidths)
//    //    GammaL: specifies broadening parameter ('gamma_air' or 'gamma_self')
//    //    HITRAN_units: use cm2/molecule (True) or cm-1 (False) for absorption coefficient
//    //    File: write output to file (if specified)
//    //    Format: c-format of file output (accounts significant digits in OmegaStep)
//    // OUTPUT PARAMETERS:
//    //    Omegas: wavenumber grid with respect to parameters OmegaRange and OmegaStep
//    //    Xsect: absorption coefficient calculated on the grid.
//    //    Units are switched by HITRAN_units

//    QMessageBox msgBox;    // Message box to display information to the user
//    int numPoints;    // Number of wavenumber points
//    double tempRef;   // Reference temperature
//    double totalPressRef;  // Reference pressure
//    double factor;    // Volume concentration multiplication factor
//    double lineIntensity;  // Calculated line intensity
//    double gammaDDB;  // Gamma D DB
//    double gammaD;    // Gamma D
//    double gamma0;    // Gamma 0
//    double gamma2;    // Gamma 2
//    double gamma0DB;  // Gamma 0 DB
//    double gamma2DB;  // Gamma 2 DB
//    double shift0;    // Shift 0
//    double shift2;    // Shift 2
//    double shift0DB;  // Shift 0 DB
//    double shift2DB;  // Shift 2 DB
//    double anuVC;     // anuVC
//    double anuVCDB;   // anuVC DB
//    double wnWingF;   // wnWingF
//    int boundIndexLower;  // Lower bounding index
//    int boundIndexUpper;  // Upper bounding index
//    double cMassMol;  //
//    double m;   //

//    // Partition functions
//    double sigmaT;    // TIPS at the current temperature
//    double sigmaTRef; // TIPS at the reference temperature


//    // Reference temperature and pressure setting, the actual ones are passed into the function
//    tempRef = 296;
//    totalPressRef = 1.0;   // here the value is in mbar, or 1.0 Atm;  // The pressure values to be supplied to this routine need to be converted to atm from milibar at some point
//    //  T = Environment['T'] # K
//    //  p = Environment['p'] # atm

//    // The difference between the 'ABUNDANCE' and the 'NATURAL ABUNDANCE' ?

//    // Precalculation of volume concentration
//    // Depending on the HITRAN database units taking into account the multiplication factor
//    if (hitranUnits == true)
//    {
//        // Multiplication factor is equal to 1.0
//        factor = 1.0;
//    }
//    else
//    {
//        // Multiplication factor calculated from the pressure and temperature
//        factor = volumeConcentration(totalPress, temp);
//    }

//    // Setting the broadening
//    if (par->getLineBroadening() == 0)  // Air
//    { gamma0DB = rec->getAirHw(); }
//    if (par->getLineBroadening() == 1)  // Self
//    { gamma0DB = rec->getSelfHw(); }
//    if (par->getLineBroadening() == 2)  // Both air and self
//    { gamma0DB = rec->getAirHw() + rec->getSelfHw(); }

//    // Pressure shift, taking parameters from the HITRAN database if the line shift is used
//    if (useLineShift) { shift0DB = rec->getAirPShift(); }
//                 else { shift0DB = 0; }

//    // If the abundance is unavailable, then skipping the current calculation
//    if (par->getMolAt(rec->getMolNum() - 1)->getIsoAt(rec->getIsoNum() - 1)->getAbund() == 0) { return false; }

//    // Partition functions for T and Tref, need to be optimized
//    sigmaT = par->getMolAt(rec->getMolNum() - 1)->getIsoAt(rec->getIsoNum() - 1)->getTips()->at(temp - 70);   // partitionFunction(MoleculeNumberDB,IsoNumberDB,T)
//    sigmaTRef = par->getMolAt(rec->getMolNum() - 1)->getIsoAt(rec->getIsoNum() - 1)->getQ296K();   // partitionFunction(MoleculeNumberDB,IsoNumberDB,Tref)

//    // Getting all environment dependences from Voigt parameters, line intensity
//    lineIntensity = environmentDependencyIntensity(rec->getIntens(),
//                                                   temp,
//                                                   tempRef,
//                                                   sigmaT,
//                                                   sigmaTRef,
//                                                   rec->getLowE(),
//                                                   rec->getWn());

//    // Filtering by LineIntensity: compare it with IntencityThreshold. Need to apply wing narrowing instead of filtering, this would be more appropriate
//    if (lineIntensity < intensThresh) { return false; }  // Skip this line then, not included

//    // Doppler broadening coefficient (GammaD), V1 (need to find out where cc comes from)
//    gammaDDB = Const::cSqrtLn2 * rec->getWn() / Const::cc * sqrt((double) 2 * Const::cBolts * temp / par->getMolAt(rec->getMolNum() - 1)->getIsoAt(rec->getIsoNum() - 1)->getMolarMass());

//    // GammaD = EnvironmentDependency_GammaD(GammaDDB,T,Tref), V2
//    cMassMol = 1.66053873e-27;  // hapi # cMassMol = 1.6605402e-27 # converter
//    m = par->getMolAt(rec->getMolNum() - 1)->getIsoAt(rec->getIsoNum() - 1)->getMolarMass() * cMassMol * 1000;
//    gammaD = sqrt(2 * Const::cBolts * temp * log(2) / m / (Const::cc * Const::cc)) * rec->getWn();

//    // Lorentz broadening coefficient
//    gamma0 = environmentDependencyGamma0(gamma0DB,
//                                         temp,
//                                         tempRef,
//                                         totalPress,
//                                         totalPressRef,
//                                         rec->getNAir());   // tempRatioPowerDB -> make sure this corresponds to the NAir in the HITRAN database

//    // Quadratic speed dependence of lorentz broadening coefficient
//    // gamma2DB is a parameter only available for the Hartman-Trann profile from what I understand, needs to be specified in addition to the data available in the HITRAN database I think
//    gamma2 = gamma2DB * totalPress / totalPressRef * pow((double) (tempRef / temp), (double) rec->getNAir());

//    // Shift coefficient
//    shift0 = shift0DB * totalPress / totalPressRef;

//    // Quadratic speed dependence of shift coefficient
//    // This is probably also Hartmann-Tran profiile only related calculation parameter
//    shift2 = shift2DB * totalPress / totalPressRef;

//    // Dicke narrowing coefficient
//    // This also seems to be a parameter which needs to be specified specifically for the Hartmann-Tran profile
//    anuVC = anuVCDB * (totalPress / totalPressRef) * (tempRef / temp);

//    // Get final wing of the line according to Gamma0, OmegaWingHW and OmegaWing.
//    // OmegaWingF = max(OmegaWing, OmegaWingHW * Gamma0, OmegaWingHW * GammaD);
//    // Will do for now but this should probably be replaced with a bubble sort or something
//    if ((wnWingAbs > (wnWingStHw * gamma0)) && (wnWingAbs > (wnWingStHw * gammaD))) { wnWingF = wnWingAbs;; }
//    if (((wnWingStHw * gamma0) > wnWingAbs) && ((wnWingStHw * gamma0) > (wnWingStHw * gammaD))) { wnWingF = (wnWingStHw * gamma0); }
//    if (((wnWingStHw * gammaD) > wnWingAbs) && ((wnWingStHw * gammaD) > (wnWingStHw * gamma0))) { wnWingF = (wnWingStHw * gammaD); }

//    // PROFILE_VOIGT(sg0, GamD, Gam0, sg)
//    //     sg0 : Unperturbed line position in cm-1 (Input).
//    //    GamD : Doppler HWHM in cm-1 (Input)
//    //    Gam0 : Speed-averaged line-width in cm-1 (Input).
//    //      sg : Current WaveNumber of the Computation in cm-1 (Input).

//    // Determining the lower and upper boundary indexes for a single spectral line based on the wing contribution for a single line
//    boundIndexLower = wavelength->getWns()->count() * ((rec->getWn() - wnWingF - wavelength->getStartWn()) / (wavelength->getEndWn() - wavelength->getStartWn()));
//    boundIndexUpper = wavelength->getWns()->count() * ((rec->getWn() + wnWingF - wavelength->getStartWn()) / (wavelength->getEndWn() - wavelength->getStartWn()));
//    spec->setStartIndex(boundIndexLower);
//    spec->setEndWn(boundIndexUpper);
//    spec->getY()->resize(boundIndexUpper - boundIndexLower + 1);


//    // Lineshape values:
//    // lineshape_vals = PROFILE_HT(LineCenterDB, GammaD, Gamma0, Omegas[BoundIndexLower:BoundIndexUpper])[0];
//    // lineshape_vals = PROFILE_HT(LineCenterDB,
//    //                             GammaD,
//    //                             Gamma0,
//    //                             Gamma2,
//    //                             Shift0,
//    //                             Shift2,
//    //                             anuVC,
//    //                             eta,
//    //                             Omegas[BoundIndexLower:BoundIndexUpper])[0]
//    //                             Xsect[BoundIndexLower:BoundIndexUpper] += factor
//    //                             NATURAL_ABUNDANCES[(MoleculeNumberDB,IsoNumberDB)]

//    // if File: save_to_file(File, Format, Omegas, Xsect)
//    //    return Omegas, Xsect;   // In our case the function should simply save the data to the harddrive;

//    // Saving the obtained calculated coefficient for this given spectral line

//    // Returning true indicating that the calculation for this line was successfull
//      return true;
//}


// Volume concentration of all gas molecules at pressure "press" and temperature "temp", (HAPI converted)
// This method is converted from the Harvard Smithsonian's HAPI implementation
//double Lbl::volumeConcentration(double press, double temp)
//{
//    return (press /9.869233e-7) / (Const::cBolts * temp);   // CGS
//}

// Temperature dependence for intensities (HITRAN), (HAPI converted)
// This method is converted from the Harvard Smithsonian's HAPI implementation
//double Lbl::environmentDependencyIntensity(double lineIntensityRef,
//                                          double temp,
//                                          double tempRef,
//                                          double sigmaT,
//                                          double sigmaTref,
//                                          double lowerStateEnergy,
//                                          double lineCenter)
//{
//    double lineIntensity;
//    double ch;
//    double zn;

//    double c = 1.4388028496642257;
//    ch = exp(-c * lowerStateEnergy / temp) * (1 - exp(-c * lineCenter / temp));
//    zn = exp(-c * lowerStateEnergy / tempRef)*(1 - exp(-c * lineCenter / tempRef));
//    lineIntensity = lineIntensityRef * (sigmaTref / sigmaT) * (ch / zn);
//    return lineIntensity;
//}

// Environmental dependence for Gamma0 (HTP, Voigt), (HAPI converted)
// This method is converted from the Harvard Smithsonian's HAPI implementation
//double Lbl::environmentDependencyGamma0(double gamma0Ref,
//                                        double temp,
//                                        double tempRef,
//                                        double press,
//                                        double pressRef,
//                                        double tempRatioPower)
//{
//    return gamma0Ref * (press / pressRef) * pow((double) (tempRef / temp), (double) tempRatioPower);
//}

// Determine default parameters from those which are passed to absorptionCoefficient_... routine
// This method is converted from the Harvard Smithsonian's HAPI implementation
//void getDefaultValuesForXsect(Params *par,
//                              HitranOneRec *rec,
//                              double press,
//                              double temp,
//                              double wnMin,
//                              double wnMax,
//                              double wnStep,
//                              double wnWing,
//                              double intensityThreshold,
//                              bool format)
//{
//    ;
//}

// Function to calculate the Hartmann-Tran profile
// This method is converted from the Harvard Smithsonian's HAPI implementation
//void Lbl::ProfileHT(double sg0,
//                    double GamD,
//                    double Gam0,
//                    double Gam2,
//                    double Shift0,
//                    double Shift2,
//                    double anuVC,
//                    double eta,
//                    double sg)
//{
    //  Input/Output Parameters of Routine (Arguments or Common)
    //  -----------------------------------------------------------
    //  T      -> Temperature in Kelvin (Input).
    //  amM1   -> Molar mass of the absorber in g/mol(Input).
    //  sg0    -> Unperturbed line position in cm-1 (Input).
    //  GamD   -> Doppler HWHM in cm-1 (Input)
    //  Gam0   -> Speed-averaged line-width in cm-1 (Input).
    //  Gam2   -> Speed dependence of the line-width in cm-1 (Input).
    //  anuVC  -> Velocity-changing frequency in cm-1 (Input).
    //  eta    -> Correlation parameter, No unit (Input).
    //  Shift0 -> Speed-averaged line-shift in cm-1 (Input).
    //  Shift2 -> Speed dependence of the line-shift in cm-1 (Input)
    //  sg     -> Current WaveNumber of the Computation in cm-1 (Input).

    // -------------------------------------------------
    //  "pCqSDHC": partially-Correlated Quadratic-Speed-Dependent Hard-Collision
    //  Subroutine to Compute the complex normalized spectral shape of an
    //  isolated line by the pCqSDHC model
    //
    //  References:
    //
    //  1) N.H. Ngo, D. Lisak, H. Tran, J.-M. Hartmann.
    //     An isolated line-shape model to go beyond the Voigt profile in
    //     spectroscopic databases and radiative transfer codes.
    //     JQSRT, Volume 129, November 2013, Pages 89–100
    //     http://dx.doi.org/10.1016/j.jqsrt.2013.05.034
    //
    //  2) H. Tran, N.H. Ngo, J.-M. Hartmann.
    //     Efficient computation of some speed-dependent isolated line profiles.
    //     JQSRT, Volume 129, November 2013, Pages 199–203
    //     http://dx.doi.org/10.1016/j.jqsrt.2013.06.015
    //
    //  3) H. Tran, N.H. Ngo, J.-M. Hartmann.
    //     Erratum to “Efficient computation of some speed-dependent isolated line profiles”.
    //     JQSRT, Volume 134, February 2014, Pages 104
    //     http://dx.doi.org/10.1016/j.jqsrt.2013.10.015
    //
    //  Input/Output Parameters of Routine (Arguments or Common)
    //  ---------------------------------
    //     T      -> Temperature in Kelvin (Input).
    //     amM1   -> Molar mass of the absorber in g/mol(Input).
    //     sg0    -> Unperturbed line position in cm-1 (Input).
    //     GamD   -> Doppler HWHM in cm-1 (Input)
    //     Gam0   -> Speed-averaged line-width in cm-1 (Input).
    //     Gam2   -> Speed dependence of the line-width in cm-1 (Input).
    //     anuVC  -> Velocity-changing frequency in cm-1 (Input).
    //     eta    -> Correlation parameter, No unit (Input).
    //     Shift0 -> Speed-averaged line-shift in cm-1 (Input).
    //     Shift2 -> Speed dependence of the line-shift in cm-1 (Input)
    //     sg     -> Current WaveNumber of the Computation in cm-1 (Input).
    //
    //  The function has two outputs:
    //  ----------------------------------
    //     (1): Real part of the normalized spectral shape (cm)
    //     (2): Imaginary part of the normalized spectral shape (cm)
    //
    //     Called Routines: 'CPF'       (Complex Probability Function)
    //     ---------------  'CPF3'      (Complex Probability Function for the region 3)
    //
    //     Based on a double precision Fortran version
    //
    //  ----------------------------------
    //

    // return pcqsdhc(sg0, GamD, Gam0, Gam2, Shift0, Shift2, anuVC, eta, sg);

    // PROFILE_HTP = PROFILE_HT - Stub for backwards compatibility
// }


// ---------------------- Hartmann-Tran Profile (HTP) ---------------------------
// This method is converted from the Harvard Smithsonian's HAPI implementation with modifications to the C++ program strucure etc.
//void Lbl::pcqsdhc(double sg0,
//                  double GamD,
//                  double Gam0,
//                  double Gam2,
//                  double Shift0,
//                  double Shift2,
//                  double anuVC,
//                  double eta,
//                  QVector<double> sg,
//                  int startIndex,
//                  int endIndex)
//{

// Example of setting the Hartmann-Tran profile
//     w0 = 1000.
//     GammaD = 0.005
//     Gamma0 = 0.2
//     Gamma2 = 0.01 * Gamma0
//     Delta0 = 0.002
//     Delta2 = 0.001 * Delta0
//     nuVC = 0.2
//     eta = 0.5
//     Dw = 1.
//     ww = arange(w0-Dw, w0+Dw, 0.01) # GRID WITH THE STEP 0.01
//     l1 = PROFILE_HT(w0,GammaD,Gamma0,Gamma2,Delta0,Delta2,nuVC,eta,ww)[0]  ;  l1 contains values of HT profile calculates on the grid ww

// ------------------------------------------------------------------------------
//     "pCqSDHC": partially-Correlated quadratic-Speed-Dependent Hard-Collision
//     Subroutine to Compute the complex normalized spectral shape of an
//     isolated line by the pCqSDHC model
//
//     Reference:
//     H. Tran, N.H. Ngo, J.-M. Hartmann.
//     Efficient computation of some speed-dependent isolated line profiles.
//     JQSRT, Volume 129, November 2013, Pages 199–203
//     http://dx.doi.org/10.1016/j.jqsrt.2013.06.015
//
//     Input/Output Parameters to be passed to a routine (Arguments or Common)
// ------------------------------------------------------------------------------
//       T      -> Temperature in Kelvin (Input).
//       amM1   -> Molar mass of the absorber in g/mol(Input).
//       sg0    -> Unperturbed line position in cm-1 (Input).
//       GamD   -> Doppler HWHM in cm-1 (Input)
//       Gam0   -> Speed-averaged line-width in cm-1 (Input).
//       Gam2   -> Speed dependence of the line-width in cm-1 (Input).
//       anuVC  -> Velocity-changing frequency in cm-1 (Input).
//       eta    -> Correlation parameter, No unit (Input).
//       Shift0 -> Speed-averaged line-shift in cm-1 (Input).
//       Shift2 -> Speed dependence of the line-shift in cm-1 (Input)
//       sg     -> Current WaveNumber of the Computation in cm-1 (Input).
//
//     Output Quantities (through Common Statements)
// ------------------------------------------------------------------------------
//       LS_pCqSDHC_R: Real part of the normalized spectral shape (cm)
//       LS_pCqSDHC_I: Imaginary part of the normalized spectral shape (cm)
//
//       Called Routines: 'CPF'    (Complex Probability Function)
//       ---------------  'CPF3'   (Complex Probability Function for the region 3)
//
//       Called By: Main Program
//       ---------------
//
//       Double Precision Version
//
// ------------------------------------------------------------------------------

//    double cte;  // Complex variable
//    double rpi;  // Complex variable
//    std::complex<double> iz;   // Complex variable
//    std::complex<double> c0;   // Complex variable
//    std::complex<double> c2;   // Complex variable
//    std::complex<double> c0t;  // Complex variable
//    std::complex<double> c2t;  // Complex variable
//    QVector <std::complex<double> > Aterm_GLOBAL;  // Complex variable array
//    QVector <std::complex<double> > Bterm_GLOBAL;  // Complex variable array
//    QVector <std::complex<double> > X;     // Complex variable array
//    std::complex<double> Y;    // Complex scalar number
//    QVector <std::complex<double> > Z1;    // Complex variable array
//    QVector <std::complex<double> > Z2;    // Complex variable array
//    QVector <double> xZ1;  // Vriable array
//    QVector <double> yZ1;  // Variable array
//    QVector <double> WR1;   // Complex variable array
//    QVector <double> WI1;   // Complex variable array
//    QVector <double> WR2;   // Complex variable array
//    QVector <double> WI2;   // Complex variable array
//    QVector <std::complex<double> > X_TMP; // Temporary complex array
//    QVector <double> xZ2;  // Variable array
//    QVector <double> yZ2;  // Variable array
//    QVector <double> SZ1;  // Variable array
//    QVector <double> SZ2;  // Variable array
//    QVector <double> DSZ;  // Variable array
//    QVector <double> WR1_PART4;   // WR1_PART4 = zeros(length_PART4)  // Probably use append for these arrays to increment them at each successive cycle
//    QVector <double> WI1_PART4;   // WI1_PART4 = zeros(length_PART4)
//    QVector <double> WR2_PART4;   // WR2_PART4 = zeros(length_PART4)
//    QVector <double> WI2_PART4;   // WI2_PART4 = zeros(length_PART4)
//    QVector <double> WR1_PART2;   // WR1_PART2 -> Probably use append for these arrays to increment them at each successive cycle
//    QVector <double> WI1_PART2;   // WI1_PART2
//    QVector <double> WR2_PART2;   // WR2_PART2
//    QVector <double> WI2_PART2;   // WI2_PART2
//    QVector <double> WR1_PART3;   // WR1_PART3 -> Probably use append for these arrays to increment them at each successive cycle
//    QVector <double> WI1_PART3;   // WI1_PART3
//    QVector <double> WR2_PART3;   // WR2_PART3
//    QVector <double> WI2_PART3;   // WI2_PART3
//    std::complex<double> auxComplex;  // Auxiliary complex number
//    std::complex<double> csqrtY;  // Double auxiliary variable
//    std::complex<double> LS_pCqSDHC;  // Complex variable to return the results of the calculations from a function
//    int numPoints;  // Variable containing the number of points
//    int i;   // Loop variable
//    int si;  // Wavenumber index symmetrical relative to the center of the line
//    bool index_PART2;  // Boolean variable indicating if a condition is satisfied
//    bool index_PART3;  // Boolean variable indicating if a condition is satisfied
//    bool index_PART4;  // Boolean variable indicating if a condition is satisfied
//    double SZmx;  // Maximum value
//    double SZmn;  // Minimum value
//    bool index_CPF3;  // Boolean variable indicating if a condition is satisfied for a given index
//    bool index_CPF;   // Boolean variable indicating if a condition for a n index is met
//    bool index_ABS;   // Boolean variable indicating if an absolute value comparison is valid
//    bool index_NOT_ABS;  // Boolean variable to store the result of a conditional statement
//    double WRb;  // Auxiliary complex variable, not sure what this one does
//    double WIb;  // Auxiliary complex variable, not sure what this one does
//    double xXb;  // Auxiliary double variable
//    double yXb;  // Auxiliary double variable


//    // SG is the only vector argument which is passed to function
//    // SG is basically an array of wavenumbers to run the calculations for

//    // If type(sg) not in set([array, ndarray, list, tuple]):
//    // sg = array([sg])

//    // Saving the number of points in the array
//    numPoints = endIndex - startIndex;

//    // Aterm_GLOBAL = zeros(number_of_points,dtype=__ComplexType__)
//    // Bterm_GLOBAL = zeros(number_of_points,dtype=__ComplexType__)
//    Aterm_GLOBAL.resize(numPoints);
//    // Initializing the real and imaginary part of the number
//    for (i = 0; i < Aterm_GLOBAL.count(); i++)
//    {
//        Aterm_GLOBAL[i].real(0);
//        Aterm_GLOBAL[i].imag(0);
//    }
//    Bterm_GLOBAL.resize(numPoints);
//    for (i = 0; i < Bterm_GLOBAL.count(); i++)
//    {
//        Bterm_GLOBAL[i].real(0);
//        Bterm_GLOBAL[i].imag(0);
//    }

//    // Resizing arrays used in the calculations
//    X.resize(numPoints);
//    Z1.resize(numPoints);
//    xZ1.resize(numPoints);
//    yZ1.resize(numPoints);
//    X_TMP.resize(numPoints);
//    Z1.resize(numPoints);
//    Z2.resize(numPoints);
//    xZ1.resize(numPoints);
//    yZ1.resize(numPoints);
//    xZ2.resize(numPoints);
//    yZ2.resize(numPoints);
//    SZ1.resize(numPoints);
//    SZ2.resize(numPoints);
//    DSZ.resize(numPoints);
//    WR1.resize(numPoints);
//    WI1.resize(numPoints);

//    // Setting values of variables (including complex variables)
//    cte = sqrt((double) log((double) 2.0e0)) / GamD;
//    rpi = sqrt((double) Const::pi);

//    // iz = __ComplexType__(0.0e0 + 1.0e0j);  // Initialize this complex number
//    iz.real(0.0e0);
//    iz.imag(1.0e0);

//    // c0 = __ComplexType__(Gam0 + 1.0e0j*Shift0)
//    c0.real(Gam0);
//    c0.imag(Shift0);

//    // c2 = __ComplexType__(Gam2 + 1.0e0j*Shift2)
//    c2.real(Gam2);
//    c2.imag(Shift2);

//    // c0t = __ComplexType__((1.0e0 - eta) * (c0 - 1.5e0 * c2) + anuVC)
//    c0t = (1.0e0 - eta) * (c0 - 1.5e0 * c2) + anuVC;

//    // c2t = __ComplexType__((1.0e0 - eta) * c2)
//    c2t = (1.0e0 - eta) * c2;

//    // PART 1 of the algorithm
//    if (std::abs(c2t) == 0.0e0)
//    {

//        for (i = startIndex; i < endIndex; i++)
//        {
//            Z1[i - startIndex] = (iz * (sg0 - sg[i]) + c0t) * cte;
//            xZ1[i - startIndex] = (-1) * (Z1[i - startIndex]).imag();  // This is supposed to be an array judging by the cpf() procedure
//            yZ1[i - startIndex] = (Z1[i - startIndex]).real();   // This is supposed to be an array judging by the cpf() procedure
//        }
//        // Need to implement this function
//        cpf(&xZ1, &yZ1, &WR1, &WI1);

//        // Running the spectral line calculation over the entire wavenumber span
//        for (i = 0; i < Aterm_GLOBAL.count(); i++)
//        {
//            auxComplex.real(0);
//            auxComplex.imag(i);
//            Aterm_GLOBAL[i] = rpi * cte * (WR1[i] + auxComplex * WI1[i]);
//            for (i = 0; i < Z1.count(); i++)
//            {
//                if (std::abs(Z1[i]) <= 4.0e3)
//                {
//                    // Calculation for this i
//                    // Bterm_GLOBAL = rpi*cte*((1.0e0 - Z1**2)*__ComplexType__(WR1 + 1.0e0j*WI1) + Z1/rpi)
//                    Bterm_GLOBAL[i] = rpi * cte * ((1.0e0 - Z1[i] * Z1[i]) * (WR1[i] + auxComplex * WI1[i]) + Z1[i] / rpi);
//                }
//                else  // Not sure if this is correct or the symmetrical approach
//                {
//                    // Bterm_GLOBAL = cte*(rpi*__ComplexType__(WR1 + 1.0e0j*WI1) + 0.5e0/Z1 - 0.75e0/(Z1**3))
//                    Bterm_GLOBAL[si] = cte * (rpi * (WR1[si] + auxComplex * WI1[si]) + 0.5e0 / Z1[si] - 0.75e0 / (Z1[si] * Z1[si] * Z1[si]));
//                }
//            }
//        }
//    }
//    else   // Part 2, Part 3 and Part 4   (Part 4 is the main part)
//    {
//        // X - vector, Y - scalar
//        for (i = startIndex; i < endIndex; i++)
//        {
//            X[i - startIndex] = (iz * (sg0 - sg[i]) + c0t) / c2t;
//        }
//        Y = (1.0e0 / ((2.0e0 * cte * c2t) * (2.0e0 * cte * c2t)));
//        csqrtY = (Gam2 - iz * Shift2) / (2.0e0 * cte * (1.0e0 - eta) * (Gam2 * Gam2 + Shift2 * Shift2));

//        // PART 4

//        // These indices are probably arrays
//        for (i = 0; i < X.count(); i++)
//        {
//            index_PART2 = abs(X[i]) <= (3.0e-8 * abs(Y));
//            index_PART3 = (abs(Y) <= (1.0e-15 * abs(X[i]))) && (!index_PART2);  // Not sure how to correctly convert the "-" and "&" here what type of operation does it do? Opposite direction or logical "Not"
//            index_PART4 = !(index_PART2 || index_PART3);  // Not sure how to properly convert this one into C++, need to figure this out, the "|" operator meaning in this case

//            // Part 4
//            if (index_PART4)
//            {
//                // sqrt of a complex number, figure out how to handle this case

//                // These arrays need to be resized prior to this operation or else an append function used on them instead
//                X_TMP[i] = X[i];   // X_TMP = X[index_PART4]
//                Z1[i] = sqrt(X_TMP[i] + Y) - csqrtY;   // Z1 = sqrt(X_TMP + Y) - csqrtY
//                Z2[i] = Z1[i] + (2.0e0) * csqrtY;      // Z2 = Z1 + __FloatType__(2.0e0) * csqrtY
//                xZ1[i] = (-1) * Z1[i].imag();   // xZ1 = -Z1.imag
//                yZ1[i] =  Z1[i].real();         // yZ1 = Z1.real
//                xZ2[i] = (-1) * Z2[i].imag();   // xZ2 = -Z2.imag
//                yZ2[i] = Z2[i].real();          // yZ2 = Z2.real
//                SZ1[i] = sqrt(xZ1[i] * xZ1[i] + yZ1[i] * yZ1[i]);  // SZ1 = sqrt(xZ1**2 + yZ1**2)
//                SZ2[i] = sqrt(xZ2[i] * xZ2[i] + yZ2[i] * 2);  // SZ2 = sqrt(xZ2**2 + yZ2**2)
//                DSZ[i] = abs(SZ1[i] - SZ2[i]);   // DSZ = abs(SZ1 - SZ2)

//                //   SZmx = maximum(SZ1, SZ2)   // Check if this is complex or not. Need to reimplement the maximum function (This is a Python internal function)
//                //   SZmn = minimum(SZ1, SZ2)   // Check if this is complex or not. Need to reimplement the minimum function (This is a Python internal function)
//                //   length_PART4 = len(index_PART4)   // This is the number of indices complying with this condition, not sure how to implement this correctly
//                WR1_PART4.append(0);   // WR1_PART4 = zeros(length_PART4)  // Probably use append for these arrays to increment them at each successive cycle
//                WI1_PART4.append(0);   // WI1_PART4 = zeros(length_PART4)
//                WR2_PART4.append(0);   // WR2_PART4 = zeros(length_PART4)
//                WI2_PART4.append(0);   // WI2_PART4 = zeros(length_PART4)

//                index_CPF3 = (DSZ[i] <= 1.0e0) && (SZmx > 8.0e0) && (SZmn <= 8.0e0);
//                index_CPF = !index_CPF3;   // HAPI comment says this can be removed, but why?, need to determine what "-" does
//                if (index_CPF3)
//                {
//                    cpf3(xZ1[i], yZ1[i], &(WR1[i]), &(WI1[i]));   // WR1,WI1 = cpf3(xZ1[index_CPF3],yZ1[index_CPF3])
//                    cpf3(xZ2[i], yZ2[i], &(WR2[i]), &(WI2[i]));   // WR2,WI2 = cpf3(xZ2[index_CPF3],yZ2[index_CPF3])
//                    WR1_PART4.append(WR1[i]);   // WR1_PART4[index_CPF3] = WR1
//                    WI1_PART4.append(WI1[i]);   // WI1_PART4[index_CPF3] = WI1
//                    WR2_PART4.append(WR2[i]);   // WR2_PART4[index_CPF3] = WR2
//                    WI2_PART4.append(WI2[i]);   // WI2_PART4[index_CPF3] = WI2
//                }
//                if (index_CPF)
//                {
//                    cpf(xZ1[i], yZ1[i], &WR1[i], &WI1[i]);   // WR1,WI1 = cpf(xZ1[index_CPF],yZ1[index_CPF])
//                    cpf(xZ2[i], yZ2[i], &WR2[i], &WI2[i]);   // WR2,WI2 = cpf(xZ2[index_CPF],yZ2[index_CPF])
//                    WR1_PART4.append(WR1[i]);   // WR1_PART4[index_CPF] = WR1
//                    WI1_PART4.append(WI1[i]);   // WI1_PART4[index_CPF] = WI1
//                    WR2_PART4.append(WR2[i]);   // WR2_PART4[index_CPF] = WR2
//                    WI2_PART4.append(WI2[i]);   // WI2_PART4[index_CPF] = WI2
//                }

//                Aterm_GLOBAL[i] = rpi * cte * ((WR1_PART4[i] + auxComplex * WI1_PART4[i]) - (WR2_PART4[i] + auxComplex * WI2_PART4[i]));   // Aterm = rpi*cte*(__ComplexType__(WR1_PART4 + 1.0e0j*WI1_PART4) - __ComplexType__(WR2_PART4+1.0e0j*WI2_PART4))
//                Bterm_GLOBAL[i] = (-1.0e0 + rpi / (2.0e0 * csqrtY) * (1.0e0 - Z1[i] * Z1[i]) * (WR1_PART4[i] + auxComplex * WI1_PART4[i])- rpi / (2.0e0 * csqrtY) * (1.0e0 - Z2[i] * Z2[i]) * (WR2_PART4[i] + auxComplex * WI2_PART4[i])) / c2t;  // Bterm = (-1.0e0 + rpi/(2.0e0*csqrtY)*(1.0e0 - Z1**2)*__ComplexType__(WR1_PART4 + 1.0e0j*WI1_PART4)- rpi/(2.0e0*csqrtY)*(1.0e0 - Z2**2)*__ComplexType__(WR2_PART4 + 1.0e0j*WI2_PART4)) / c2t
//                // Aterm_GLOBAL[index_PART4] = Aterm  -> Part 4 should be taken into account
//                // Bterm_GLOBAL[index_PART4] = Bterm  -> Part 4 should be taken into account
//            }

//            // Part 2
//            if (index_PART2)
//            {
//                X_TMP[i] = X[i];  // X_TMP = X[index_PART2]
//                Z1[i] = (iz * (sg0 - sg[i]) + c0t) * cte;   // Z1 = (iz*(sg0 - sg[index_PART2]) + c0t) * cte
//                // Z2 = sqrt(X_TMP[i] + Y[i]) + csqrtY;  // Z2 = sqrt(X_TMP + Y) + csqrtY, not sure how to handle this, seems like a complex number
//                xZ1[i] = (-1) * Z1[i].imag();   // xZ1 = -Z1.imag
//                yZ1[i] = Z1[i].real();          // yZ1 = Z1.real
//                xZ2[i] = (-1) * Z2[i].imag();   // xZ2 = -Z2.imag
//                yZ2[i] = Z2[i].real();          // yZ2 = Z2.real
//                cpf(xZ1[i], yZ1[i], &WR1_PART2[i], &WI1_PART2[i]);  // WR1_PART2, WI1_PART2 = cpf(xZ1,yZ1)
//                cpf(xZ2[i], yZ2[i], &WR2_PART2[i], &WI2_PART2[i]);  // WR2_PART2, WI2_PART2 = cpf(xZ2,yZ2)
//                Aterm_GLOBAL[i] = rpi * cte * ((WR1_PART2[i] + auxComplex * WI1_PART2[i]) - (WR2_PART2[i] + auxComplex * WI2_PART2[i]));
//                Bterm_GLOBAL[i] = (-1.0e0 + rpi / (2.0e0 * csqrtY) * (1.0e0 - Z1[i] * Z1[i]) * (WR1_PART2[i] + auxComplex * WI1_PART2[i]) - rpi / (2.0e0 * csqrtY) * (1.0e0 - Z2[i] *Z2[i]) * (WR2_PART2[i] + auxComplex * WI2_PART2[i])) / c2t;
//                //      Aterm_GLOBAL[index_PART2] = Aterm  -> Part 2 should be taken into account
//                //      Bterm_GLOBAL[index_PART2] = Bterm  -> Part 2 should be taken into account
//            }

//            // Part 3
//            if (index_PART3)
//            {
//                X_TMP[i] = X[i];  // X_TMP = X[index_PART3]
//                xZ1[i] = (-1) * (sqrt(std::complex<double>(X_TMP[i] + Y))).imag();  // xZ1 = -sqrt(X_TMP + Y).imag    // Need to figure out how to implement sqrt on complex numbers
//                yZ1[i] = sqrt(std::complex<double>(X_TMP[i] + Y)).real();   // yZ1 = sqrt(X_TMP + Y).real    // Need to figure out how to implement sqrt on complex numbers
//                cpf(xZ1[i], yZ1[i], &WR1_PART3[i], &WI1_PART3[i]);  // WR1_PART3, WI1_PART3 =  cpf(xZ1, yZ1)
//                index_ABS = abs(sqrt(X_TMP[i])) <= 4.0e3;   // index_ABS = abs(sqrt(X_TMP)) <= 4.0e3
//                index_NOT_ABS = !index_ABS;      // This operator is probably simply a negation operator or the logical Not (in either case need to figure this out)
//                Aterm_GLOBAL[i] = 0;   // Aterm = zeros(len(index_PART3), dtype=__ComplexType__)
//                Bterm_GLOBAL[i] = 0;   // Bterm = zeros(len(index_PART3), dtype=__ComplexType__)

//                if (index_ABS)
//                {
//                    xXb = -sqrt(X[i]).imag();
//                    yXb = sqrt(X[i]).real();
//                    cpf(xXb, yXb, &WRb, &WIb);
//                    Aterm_GLOBAL[i] = (2.0e0 * rpi / c2t) * (1.0e0 / rpi - sqrt(X_TMP[i]) * (WRb + auxComplex * WIb));
//                    Bterm_GLOBAL[i] = (1.0e0 / c2t) * (-1.0e0 + 2.0e0 * rpi * (1.0e0 - X_TMP[i]-2.0e0 * Y) * (1.0e0 / rpi - sqrt(X_TMP[i]) * (WRb + auxComplex * WIb)) + 2.0e0 * rpi * sqrt(std::complex<double>(X_TMP[i] + Y)) * (WR1_PART3[i] + auxComplex * WI1_PART3[i]));
//                }

//                if (index_NOT_ABS)
//                {
//                    Aterm_GLOBAL[i] = (1.0e0 / c2t) * (1.0e0 / X_TMP[i] - 1.5e0 / (X_TMP[i] * X_TMP[i]));
//                    Bterm_GLOBAL[i] = (1.0e0 / c2t) * ( -1.0e0 + (1.0e0 - X_TMP[i] - 2.0e0 * Y) * (1.0e0 / X_TMP[i] - 1.5e0 / (X_TMP[i] * X_TMP[i])) + 2.0e0 * rpi * sqrt(std::complex<double>(X_TMP[i] + Y)) * (WR1[i] + auxComplex * WI1[i]));
//                    // Aterm_GLOBAL[index_PART3] = Aterm  //
//                    // Bterm_GLOBAL[index_PART3] = Bterm  // Determine what this operation actually does with the array index where the insertion is taking place spelled out like that.
//                }
//            }
//        }
//    }

//    // Common part
//    for (i = 0; i < Aterm_GLOBAL.count(); i++)
//    {
//        LS_pCqSDHC = (1.0e0 / Const::pi) * (Aterm_GLOBAL[i] / (1.0e0 - (anuVC-eta*(c0-1.5e0 * c2)) * Aterm_GLOBAL[i] + eta * c2 * Bterm_GLOBAL[i]));
//    }
//    //    return LS_pCqSDHC.real, LS_pCqSDHC.imag
//}

// Complex probability function implementation (Humlicek)
// This method is converted from the Harvard Smithsonian's HAPI implementation
//void Lbl::cpf(QVector<double> *X, QVector<double> *Y, QVector<double> *WR1, QVector<double> *WI1)
//{
//    ;
//}

// Complex probability function implementation (Humlicek)
// This method is converted from the Harvard Smithsonian's HAPI implementation
//void Lbl::cpf(double X, double Y, double *WR1, double *WI1)
//{
//    int i;  // Loop variable
//    QVector<double> T;   // T array
//    QVector<double> U;   // U array
//    QVector<double> S;   // S array
//    std::complex<double> zi;    // __ComplexType__(0.0e0 + 1.0e0j)
//    std::complex<double> zone;  // __ComplexType__(1.0e0 + 0.0e0j)
//    std::complex<double> zm1;   //
//    std::complex<double> zm2;   //

//    std::complex<double> zsum_REGION3;
//    std::complex<double> zterm;
//    QVector<double> tt;  // __FloatType__([0.5e0,1.5e0,2.5e0,3.5e0,4.5e0,5.5e0,6.5e0,7.5e0,8.5e0,9.5e0,10.5e0,11.5e0,12.5e0,13.5e0,14.5e0])
//    double pipwoeronehalf = 0.564189583547756e0;
//    double WR;  // Real component of a number
//    double WI;  // Imaginary component of a number
//    // Region 1 variables
//    double WR_REGION1;  // Real part of auxiliary complex number
//    double WI_REGION1;  // Imaginary part of auxiliary complex number
//    double R_REGION1;   //
//    double X_REGION1;   //
//    double D_REGION1;   //
//    double D1_REGION1;  //
//    double D2_REGION1;  //
//    double D3_REGION1;  //
//    double D4_REGION1;  //
//    double Y1_REGION1;  //
//    double Y2_REGION1;  //
//    // Region 2 variables
//    double WR_REGION2;  // Real part of a complex number
//    double WI_REGION2;  // Imaginary part of a complex number
//    double X_REGION2;   //
//    double Y_REGION2;   //
//    double R_REGION2;   //
//    double R2_REGION2;  //
//    double D_REGION2;   //
//    double D1_REGION2;  //
//    double D2_REGION2;  //
//    double D3_REGION2;  //
//    double D4_REGION2;  //
//    double Y1_REGION2;  //
//    double Y2_REGION2;  //
//    double Y3_REGION2;  //
//    // Region 3 variables
//    double X_REGION3;   //
//    double Y_REGION3;   //
//    // Region 12 variables
//    double X_REGION12;  //
//    double Y_REGION12;  //
//    double Y1_REGION12; //
//    double Y2_REGION12; //
//    double WR_TOTAL;    // Real part of a complex number (total)
//    double WI_TOTAL;    // Imaginary part of a complex number (total)
//    double Y_REGION1;   //

//    bool index_REGION12;    // Boolean variable for selecting a sub-region of indexes
//    bool subindex_REGION2;  // Boolean variable for selecting a sub-region of indexes
//    bool index_REGION2;     // Boolean variable for selecting a sub-region of indexes
//    bool index_REGION3;     // Boolean variable for selecting a sub-region of indexes
//    bool ii;  // Boolean variable to get the conditions for a set of indexes
//    bool index_REGION1;     // Boolean variable for selecting a sub-region of indexes
//    bool subindex_REGION1;  // Boolean index to select a sub-region in a region


//    // See if this implementation may be simplified to take into account a single value algorithm replacement instead of working with arrays

//    // Initializing the static variables
//    zone.real(1.0e0);
//    zone.imag(0.0e0);
//    zi.real(0.0e0);
//    zi.imag(1.0e0);
//    tt = QVector<double> () << 0.5e0 << 1.5e0 << 2.5e0 << 3.5e0 << 4.5e0 << 5.5e0 << 6.5e0 << 7.5e0 << 8.5e0 << 9.5e0 << 10.5e0 << 11.5e0 << 12.5e0 << 13.5e0 << 14.5e0;

//    // __FloatType__([0.314240376e0,0.947788391e0,1.59768264e0,2.27950708e0,3.02063703e0,3.8897249e0])
//    T = QVector<double> () << 0.314240376e0 << 0.947788391e0 << 1.59768264e0 << 2.27950708e0 << 3.02063703e0 << 3.8897249e0;

//    // __FloatType__([1.01172805e0,-0.75197147e0,1.2557727e-2,1.00220082e-2,-2.42068135e-4,5.00848061e-7])
//    U = QVector<double> () << 1.01172805e0 << -0.75197147e0 << 1.2557727e-2 << 1.00220082e-2 << -2.42068135e-4 << 5.00848061e-7;

//    // __FloatType__([1.393237e0,0.231152406e0,-0.155351466e0,6.21836624e-3,9.19082986e-5,-6.27525958e-7
//    S = QVector<double> () << 1.393237e0 << 0.231152406e0 << -0.155351466e0 << 6.21836624e-3 << 9.19082986e-5 << -6.27525958e-7;

//    // This procedure has been changed to work on  scalar values instead of an array
//    if (sqrt(X * X + Y * Y) > (8.0e0))
//    {
//        X_REGION3 = X;
//        Y_REGION3 = Y;
//        zm1 = zone / (X_REGION3 + zi * Y_REGION3);
//        zm2 = zm1 * zm1;
//        zsum_REGION3 = zone;
//        zterm = zone;
//        for (i = 0; i < tt.count(); i++)  // Figure out how this works if this "for in" is correct
//        {
//            zterm = zterm * zm2 * tt[i];
//            zsum_REGION3 = zsum_REGION3 + zterm;
//        }
//        zsum_REGION3 = zsum_REGION3 * zi * zm1 * pipwoeronehalf;
//    }

//    // Figure out what this operation does
//    //   index_REGION12 = setdiff1d(array(arange(len(X))),array(index_REGION3))
//    if (index_REGION12)
//    {
//       X_REGION12 = X;
//       Y_REGION12 = Y;
//    }

//    // Initializing the WR and WI
//    WR = (0.0e0);
//    WI = (0.0e0);

//    // REGION 12
//    Y1_REGION12 = Y_REGION12 + (1.5e0);
//    Y2_REGION12 = Y1_REGION12 * Y1_REGION12;

//    // REGION 2
//    //    subindex_REGION2 = where((Y_REGION12 <= 0.85e0) &
//    //    (abs(X_REGION12) >= (18.1e0 * Y_REGION12 + 1.65e0)))
//    //    index_REGION2 = index_REGION12[subindex_REGION2]
//    if ((Y_REGION12 <= 0.85e0) && (abs(X_REGION12) >= (18.1e0 * Y_REGION12 + 1.65e0)))
//    {
//        //    X_REGION2 = X[index_REGION2]
//        //    Y_REGION2 = Y[index_REGION2]
//        //    Y1_REGION2 = Y1_REGION12[subindex_REGION2]
//        //    Y2_REGION2 = Y2_REGION12[subindex_REGION2]
//        X_REGION2 = X;
//        Y_REGION2 = X;
//        if (index_REGION12)
//        {
//            Y1_REGION2 = Y1_REGION12;
//            Y2_REGION2 = Y2_REGION12;
//        }
//        Y3_REGION2 = Y_REGION2 + (3.0e0);
//    }

//    WR_REGION2 = WR;
//    WI_REGION2 = WI;

//    //    WR_REGION2 = zeros(len(X_REGION2))
//    //    ii = abs(X_REGION2) < __FloatType__(12.0e0)
//    ii = (abs(X_REGION2) < (12.0e0));
//    if (ii)
//    {
//        //   WR_REGION2[ii] = exp(-X_REGION2[ii]**2)
//        WR_REGION2 = exp(-X_REGION2 * X_REGION2);
//    }
//    else
//    {
//        //   WR_REGION2[~ii] = WR
//        WR_REGION2 = WR;
//    }

//    for (i = 0; i < 6; i++)
//    {
//        R_REGION2 = X_REGION2 - T[i];
//        R2_REGION2 = R_REGION2 * R_REGION2;
//        D_REGION2 = (1.0e0) / (R2_REGION2 + Y2_REGION2);
//        D1_REGION2 = Y1_REGION2 * D_REGION2;
//        D2_REGION2 = R_REGION2 * D_REGION2;
//        WR_REGION2 = WR_REGION2 + Y_REGION2 * (U[i] * (R_REGION2 * D2_REGION2 - 1.5e0 * D1_REGION2) + S[i] * Y3_REGION2 * D2_REGION2) / (R2_REGION2 + 2.25e0);
//        R_REGION2 = X_REGION2 + T[i];
//        R2_REGION2 = R_REGION2 * R_REGION2;
//        D_REGION2 = (1.0e0) / (R2_REGION2 + Y2_REGION2);
//        D3_REGION2 = Y1_REGION2 * D_REGION2;
//        D4_REGION2 = R_REGION2 * D_REGION2;
//        WR_REGION2 = WR_REGION2 + Y_REGION2 * (U[i] * (R_REGION2 * D4_REGION2 - 1.5e0 * D3_REGION2) - S[i] * Y3_REGION2 * D4_REGION2) / (R2_REGION2 + 2.25e0);
//        WI_REGION2 = WI_REGION2 + U[i] * (D2_REGION2 + D4_REGION2) + S[i] * (D1_REGION2 - D3_REGION2);
//    }

//    //    // REGION3
//    //    index_REGION1 = setdiff1d(array(index_REGION12),array(index_REGION2))  // Figure out what this means
//    if (index_REGION1)
//    {
//        X_REGION1 = X;
//        Y_REGION1 = X;
//    }

//    //    subindex_REGION1 = setdiff1d(array(arange(len(index_REGION12))),array(subindex_REGION2))  // Figure out what this means
//    if (subindex_REGION1)
//    {
//    //    Y1_REGION1 = Y1_REGION12[subindex_REGION1]
//    //    Y2_REGION1 = Y2_REGION12[subindex_REGION1]
//        Y1_REGION1 = Y1_REGION12;
//        Y2_REGION1 = Y2_REGION12;
//    }

//    WR_REGION1 = WR;
//    WI_REGION1 = WI;
//    for (i = 0; i < 6; i++)
//    {
//        R_REGION1 = X_REGION1 - T[i];
//        D_REGION1 = (1.0e0) / (R_REGION1 * R_REGION1 + Y2_REGION1);
//        D1_REGION1 = Y1_REGION1 * D_REGION1;
//        D2_REGION1 = R_REGION1 * D_REGION1;
//        R_REGION1 = X_REGION1 + T[i];
//        D_REGION1 = (1.0e0) / (R_REGION1 * R_REGION1 + Y2_REGION1);
//        D3_REGION1 = Y1_REGION1 * D_REGION1;
//        D4_REGION1 = R_REGION1 * D_REGION1;
//        WR_REGION1 = WR_REGION1 + U[i] * (D1_REGION1 + D3_REGION1) - S[i] * (D2_REGION1 - D4_REGION1);
//        WI_REGION1 = WI_REGION1 + U[i] * (D2_REGION1 + D4_REGION1) + S[i] * (D1_REGION1 - D3_REGION1);
//    }

//    // Total result
//    //    WR_TOTAL = zeros(len(X))
//    //    WI_TOTAL = zeros(len(X))

//    // REGION 3
//    if (index_REGION3)
//    {
//        //   WR_TOTAL[index_REGION3] = zsum_REGION3.real
//        //   WI_TOTAL[index_REGION3] = zsum_REGION3.imag
//        WR_TOTAL = zsum_REGION3.real();
//        WI_TOTAL = zsum_REGION3.imag();
//    }
//    // REGION 2
//    if (index_REGION2)
//    {
//        //   WR_TOTAL[index_REGION2] = WR_REGION2
//        //   WI_TOTAL[index_REGION2] = WI_REGION2
//        WR_TOTAL = WR_REGION2;
//        WI_TOTAL = WI_REGION2;
//    }
//    // REGION 1
//    if (index_REGION1)
//    {
//        //   WR_TOTAL[index_REGION1] = WR_REGION1
//        //   WI_TOTAL[index_REGION1] = WI_REGION1
//        WR_TOTAL = WR_REGION1;
//        WI_TOTAL = WI_REGION1;
//    }

//    // return WR_TOTAL, WI_TOTAL, not necessary here
//}

// "naive" implementation for benchmarks
// This method is converted from the Harvard Smithsonian's HAPI implementation
//void Lbl::cpf3(QVector<double> *X, QVector<double> *Y, QVector<double> *WR1, QVector<double> *WI1)
//{
//    ;
//}

// "naive" implementation for benchmarks (with single values passed to arrays)
// This method is converted from the Harvard Smithsonian's HAPI implementation
//void Lbl::cpf3(double X, double Y, double *WR1, double *WI1)
//{
//    // Static variables hardwired into the program
//    int i;  // Loop variable
//    std::complex<double> zone;  // __ComplexType__(1.0e0 + 0.0e0j)
//    std::complex<double> zi;    // __ComplexType__(0.0e0 + 1.0e0j)
//    std::complex<double> zm1;   // Intermediate result complex variable 1
//    std::complex<double> zm2;   // Intermediate result complex variable 2
//    std::complex<double> zterm; // Intermediate auxiliary complex variable
//    std::complex<double> zsum;  // Intermediate auxiliary complex variable
//    QVector<double> tt;         // __FloatType__([0.5e0,1.5e0,2.5e0,3.5e0,4.5e0,5.5e0,6.5e0,7.5e0,8.5e0,9.5e0,10.5e0,11.5e0,12.5e0,13.5e0,14.5e0])
//    double pipwoeronehalf = 0.564189583547756e0;


//    // Initializing the static variables
//    zone.real(1.0e0);
//    zone.imag(0.0e0);
//    zi.real(0.0e0);
//    zi.imag(1.0e0);
//    tt = QVector<double> () << 0.5e0 << 1.5e0 << 2.5e0 << 3.5e0 << 4.5e0 << 5.5e0 << 6.5e0 << 7.5e0 << 8.5e0 << 9.5e0 << 10.5e0 << 11.5e0 << 12.5e0 << 13.5e0 << 14.5e0;

//    // Calculation part
//    zm1 = zone / (X + zi * Y); // Maybe redundant
//    zm2 = zm1 * zm1;
//    zsum = zone;
//    zterm = zone;
//    for (i = 0; i < tt.count(); i++)
//    {
//        zterm = zterm * zm2 * tt[i];
//        zsum = zsum + zterm;
//    }
//    zsum = zsum * zi * zm1 * pipwoeronehalf;
//    *WR1 = zsum.real();
//    *WI1 = zsum.imag();
//}
