// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | info@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// Parts of the line-by-line calculation algorithm are based on the Python code from the HITRAN Application Programming Interface (HAPI)
// R.V. Kochanov, I.E. Gordon, L.S. Rothman, P. Wcisło, C. Hill, J.S. Wilzewski, "HITRAN Application Programming Interface (HAPI):
// A comprehensive approach to working with spectroscopic data", JQSRT, Volume 177, July 2016, Pages 15–30; (hitran.org/hapi/)
//
// The Voigt calculation routine uses the algorithm described in: Abrarov, S., and Quine, B. (2015).
// "A rational approximation for efficient computation of the Voigt function in quantitative spectroscopy".
// Journal of Mathematics Research, 7(2), p 163. doi:http://dx.doi.org/10.5539/jmr.v7n2.
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// Object with all line-by-line calculation elements
#ifndef LBL_H
#define LBL_H

#include <QObject>
#include "hitranonerec.h"
#include "onelayerspectra.h"
#include "hitranonlinedownloader.h"
#include "params.h"
#include "spectrum.h"
#include "wavelength.h"
#include <complex>   // Complex numbers need to be used in the Hartmann-Tran profile
#include <cmath>
#include <QList>
#include <QVector>

class Lbl: public QObject
{
    Q_OBJECT
public:

    explicit Lbl(QObject *parent = 0, HitranOnlineDownloader *newHitran = 0);

    // Method to calculate spectra for a horizontal path
    void calculateHorizPath();

    // Method to calculate spectra for a slant path
    void calculateSlantPath();

    // Methods to get and set the HITRAN records to run calculations for
    QVector<HitranOneRec> *getHitranRecs();
    void setHitranRecs(QVector<HitranOneRec> *newHitranRecs);

    // Methods to get and set a pointer to the spectra array (this object is used to save the calculated spectra to)
    OneLayerSpectra *getHorizSpec();
    void setHorizSpec(OneLayerSpectra *newHorizSpec);

    // Method to get and set a pointer to the object with the calculation parameters
    Params *getParams();
    void setParams(Params *newParams);

    // Method to get and set an object with wavelengths
    Wavelength *getWavelength();
    void setWavelength(Wavelength *newWavelength);

    // Method to initialize the spectra (resize them according to the molparam.txt file information)
    void initHorizSpectra();

    // Method to find maximum values in the spectra
    void findMinMaxValuesInSpectra();

    // Method to pass the hitran downloader class reference to the object (used for segment reading of the data)
    void setHitranDownloader(HitranOnlineDownloader *newHitran);

    // Method is used to set a variable used to cancel the calculation
    void setCancelCalculation(bool newCancelCalculation);

signals:

    // Signal used to send notifications to the user of what is going on in the program
    void notify(QString message, double value, bool visible, bool cancelButtonVisible);

    // Signal is emitted to update the spectra
    void updateSpectra(bool updateScale, bool updateTotal);

    // Signal is emited to notify the core object that the calculation has been completed
    void calculationComplete();

private slots:

    // Slot to calculate a segment (for different lineshape profiles)
    void calcWnSegment();

    // Slot is executed when the calcultion is over
    void postCalculation();

private:

    // Method to create connections
    void createActions();

    // Method to calculate one Lorentz wavenumber segment
    void calcDopplerWnSegment();
    void calcLorentzWnSegment();
    void calcVoigtWnSegment();
    void calcVoigtWnSegmentHighRes();

    // Method to perform the line-by-line calculation
    void performLblCalculation();

    // The comented out portions below are for the Hartmann-Tran profile for possible implementation later
    // The implemenation has been partially completed, but not entirely

    // Method to calculate spectra for a single layer
    //    void calcOneLayerHartmannTran(QVector<HitranOneRec> *recs, OneLayerSpectra *layerSpectra);

    // Method to calculate the Hartmann-Tran absorption coefficient, (HAPI converted)
    // This method is converted from the Harvard Smithsonian's HAPI implementation
    //    bool calcSingleLineAbsorptionCoeffHartmannTran(HitranOneRec *rec,
    //                                                   int mol,
    //                                                   int iso,
    //                                                   double abund,
    //                                                   double totalPress,
    //                                                   double temp,
    //                                                   double wnStart,
    //                                                   double wnEnd,
    //                                                   double wnStep,
    //                                                   double wnWingAbs,
    //                                                   double intensThresh,
    //                                                   double wnWingStHw,
    //                                                   double broadType,
    //                                                   bool hitranUnits,
    //                                                   bool useLineShift,
    //                                                   Spectrum *spec);

    // Function to calculate the Hartmann-Tran profile
    // This method is converted from the Harvard Smithsonian's HAPI implementation
    //    void ProfileHT(double sg0,
    //                   double GamD,
    //                   double Gam0,
    //                   double Gam2,
    //                   double Shift0,
    //                   double Shift2,
    //                   double anuVC,
    //                   double eta,
    //                   double sg);

    // Method to calculate the Hartmann-Tran Profile (HTP)
    // This method is converted from the Harvard Smithsonian's HAPI implementation
    //    void pcqsdhc(double sg0,
    //                 double GamD,
    //                 double Gam0,
    //                 double Gam2,
    //                 double Shift0,
    //                 double Shift2,
    //                 double anuVC,
    //                 double eta,
    //                 QVector<double> sg,
    //                 int startIndex,
    //                 int endIndex);

    // Volume concentration of all gas molecules at pressure "press" and temperature "temp", (HAPI converted)
    // This method is converted from the Harvard Smithsonian's HAPI implementation
    //    double volumeConcentration(double press, double temp);

    // Temperature dependence for intensities (HITRAN), (HAPI converted)
    // This method is converted from the Harvard Smithsonian's HAPI implementation
    //    double environmentDependencyIntensity(double lineIntensityRef,
    //                                          double temp,
    //                                          double tempRef,
    //                                          double sigmaT,
    //                                          double sigmaTref,
    //                                          double lowerStateEnergy,
    //                                          double lineCenter);

    // Environmental dependence for Gamma0 (HTP, Voigt), (HAPI converted)
    // This method is converted from the Harvard Smithsonian's HAPI implementation
    //    double environmentDependencyGamma0(double gamma0Ref,
    //                                       double temp,
    //                                       double tempRef,
    //                                       double press,
    //                                       double pressRef,
    //                                       double tempRatioPower);

    // Determine default parameters from those which are passed to absorptionCoefficient_... routine
    // This method is converted from the Harvard Smithsonian's HAPI implementation
    //    void getDefaultValuesForXsect(Params *par,
    //                                  HitranOneRec *rec,
    //                                  double press,
    //                                  double temp,
    //                                  double wnMin,
    //                                  double wnMax,
    //                                  double wnStep,
    //                                  double wnWing,
    //                                  double intensityThreshold,
    //                                  bool format);

    // Method to get default values for cross-sections
    // This method is converted from the Harvard Smithsonian's HAPI implementation
    //    void getDefaultValuesForXsect(int mol,
    //                                  int iso,
    //                                  int abund,
    //                                  HitranOneRec *line,
    //                                  double press,
    //                                  double temp,
    //                                  double OmegaStart,
    //                                  double OmegaEnd,
    //                                  double OmegaStep,
    //                                  double OmegaWing,
    //                                  double IntensityThreshold,
    //                                  bool format);

    // Complex probability function implementation (Humlicek)
    // This method is converted from the Harvard Smithsonian's HAPI implementation
    //   void cpf(QVector<double> *X, QVector<double> *Y, QVector<double> *WR1, QVector<double> *WI1);

    // Complex probability function implementation (Humlicek)
    // This method is converted from the Harvard Smithsonian's HAPI implementation
    //   void cpf(double X, double Y, double *WR1, double *WI1);

    // "Naive" implementation for benchmarks
    // This method is converted from the Harvard Smithsonian's HAPI implementation
    //   void cpf3(QVector<double> *X, QVector<double> *Y, QVector<double> *WR1, QVector<double> *WI1);

    // "naive" implementation for benchmarks (with single values passed to arrays)
    // This method is converted from the Harvard Smithsonian's HAPI implementation
    //   void cpf3(double x, double y, double *WR1, double *WI1);


    // Array with the HITRAN data records to run calculations for
    QVector<HitranOneRec> *hitranRecs;

    // Pointer to the object with the spectra
    OneLayerSpectra *horizSpectra;

    // Pointer to the object containing the calculation parameters
    Params *par;

    // Pointer to an array with wavelengths / wavenumbers
    Wavelength *wavelength;

    // Pointer to the HITRAN downlaoder class to segment-read the HITRAN data
    HitranOnlineDownloader *hitran;

    // Current record number
    int currRecNum;

    // Variable set to cancel the calculation
    bool cancelCalculation;

    // The following coefficents used in the the Voigt profile implementation are taken from the publication reported in:
    // Abrarov, S., & Quine, B. (2015). A Rational Approximation for Efficient Computation of the Voigt Function
    // in Quantitative Spectroscopy. Journal Of Mathematics Research, 7(2), p163. doi:http://dx.doi.org/10.5539/jmr.v7n2
    // Coefficients to be used for the calculation of the Voigt profile
    QVector<double> coeff1_12;  // Alpha
    QVector<double> coeff2_12;  // Beta
    QVector<double> coeff3_12;  // Gamma
    QVector<double> coeff1_16;  // Alpha
    QVector<double> coeff2_16;  // Beta
    QVector<double> coeff3_16;  // Gamma

};

#endif // LBL_H
