// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | info@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// Object is used to relay the total pressure information betweeen the C++ data objects and the QML interface
#include "totalpressdataobject.h"
#include <QLocale>
#include <QMessageBox>

TotalPressDataObject::TotalPressDataObject(QObject *parent, BytranCore *inCore) : QObject(parent)
{
    QLocale loc;
    double totalPressAtm;        // Total pressure in default internal units (mbar)
    double totalPressCurrUnits;  // Total pressure in current pressure units


    // Passing in the core object
    core = inCore;
    totalPressAtm = core->getParams()->getHorizParams()->getTotalPress();
    totalPressCurrUnits = core->getConcPressUnitsConverter()->fromDefaultToCurrentPressUnits(totalPressAtm);
    m_totalPressure = loc.toString(totalPressCurrUnits, 'g', 6);
    m_units = core->getConcPressUnitsConverter()->getPressUnits();
}

void TotalPressDataObject::updateTotalPress()
{
    QLocale loc;
    double totalPressAtm;        // Total pressure in default internal units (mbar)
    double totalPressCurrUnits;  // Total pressure in current pressure units

    // Updating the total pressure values
    totalPressAtm = core->getParams()->getHorizParams()->getTotalPress();
    totalPressCurrUnits = core->getConcPressUnitsConverter()->fromDefaultToCurrentPressUnits(totalPressAtm);
    m_totalPressure = loc.toString(totalPressCurrUnits, 'g', 6);
    m_units = core->getConcPressUnitsConverter()->getPressUnits();
}

TotalPressDataObject::TotalPressDataObject(const QString &totalPressure, const int &units, QObject *parent)
    : QObject(parent), m_totalPressure(totalPressure), m_units(units)
{
}

QString TotalPressDataObject::totalPressure() const
{
    QLocale loc;
    double totalPressAtm;        // Total pressure in default internal units (mbar)
    double totalPressCurrUnits;  // Total pressure in current pressure units

    totalPressAtm = core->getParams()->getHorizParams()->getTotalPress();
    totalPressCurrUnits = core->getConcPressUnitsConverter()->fromDefaultToCurrentPressUnits(totalPressAtm);
    return loc.toString(totalPressCurrUnits, 'g', 6);
}

void TotalPressDataObject::setTotalPressure(const QString &totalPressure)
{
    QLocale c;  // Object to convert strings into numerical variables
    double totalPressDefaultUnits;  // Total pressure in current units to be converted to mbar


    if (totalPressure != m_totalPressure) {
        m_totalPressure = totalPressure;

        // Need to convert the value to default units (mbar)
        totalPressDefaultUnits = core->getConcPressUnitsConverter()->fromCurrentToDefaultPressUnits(c.toDouble(m_totalPressure));
        core->getParams()->getHorizParams()->setTotalPress(totalPressDefaultUnits);
        emit totalPressureChanged();
    }
}

int TotalPressDataObject::units() const
{
    return m_units;
}

void TotalPressDataObject::setUnits(const int &units)
{
    QLocale loc;
    double totalPressAtm;  // Total pressure in mbar units
    double totalPressCurrUnits;  // Total pressure in current units

    if (units != m_units) {
        m_units = units;
        // Saving the units into the object used to convert to different pathlength measurement units
        core->getConcPressUnitsConverter()->setPressUnits(m_units);

        // Updating the displayed temperature
        totalPressAtm = core->getParams()->getHorizParams()->getTotalPress();
        totalPressCurrUnits = core->getConcPressUnitsConverter()->fromDefaultToCurrentPressUnits(totalPressAtm);
        m_totalPressure = loc.toString(totalPressCurrUnits, 'g', 6);

        // Emitting signals notifying about the change in calues and units
        emit unitsChanged();
        emit totalPressureChanged();
    }    
}
